/* This file contains routines for Cholesky decomposition and
   backsubstitution used for solving linear equation sets. The
   routines are based on the formulas from 'Numerical Recipes in C,
   Second Edition, Cambridge University Press'. However, since the
   autors would not permit their code to be posted on the Internet, 
   I had to rewrite them myself. 

   Jan Kybic, October 1997 */

/* this file can be compiled separately or '#include'd (for use in MEX files)*/


#ifndef _CHOLESK_C
#define _CHOLESK_C
#include <math.h>
#include <stdlib.h>
#include <alloca.h>

/* choldc takes a positive definite and symmetric matrix a, size n*n
   and constructs its Cholesky decomposition. 'p' should point to a
   place of size n*sizeof(double), it will on output contain the
   diagonal elements. Returns zero on successful return or -1 if the
   matrix was singular or not symmetrical positive definite. Only the
   lower triangle of 'a' is changed. 
*/

int choldc(double *a,int n,double *p)
{
  int i,j,k ;
  double sum ;
  
  for(i=0;i<n;i++) {
    for(j=i;j<n;j++) {
      for(sum=a[i*n+j],k=i-1;k>=0;k--) sum -= a[i*n+k]*a[j*n+k] ;
      if (i==j) {
	if (sum<=0.0)  return -1 ;
	p[i]=sqrt(sum) ;
      } else a[j*n+i]=sum/p[i] ;
    }
  }
  return 0;
}

/* 'chol' perform Cholesky decomposition and returns the whole
   upper triangular factor in the original matrix, including the diagonal.
   The lower triangle is zeroed */
int chol(double *a,int n)
{ int i,j ;
  double *p=alloca(n*sizeof(double)) ;

  if ((i=choldc(a,n,p))!=0) return i ;

  for(i=0;i<n;i++)
    for(j=0;j<i;j++) { a[j*n+i]=a[i*n+j] ; a[i*n+j]=0.0 ; }

  for(i=0;i<n;i++) a[i*n+i]=p[i] ;
  return 0 ;
}

/* 'dechol' multiplies back the Cholesky factor from 'chol' to get the
   original matrix.  'a' is the factorised form, on output 'y' will be
   the full form */

int dechol(double *a,int n,double *y)
{ int i,j,k ;
  double sum ;

  for(i=0;i<n;i++)
    for(j=i;j<n;j++) {
      sum=0.0 ;
      for(k=0;k<n;k++) sum+=a[k*n+i]*a[k*n+j] ;
      y[i*n+j]=y[j*n+i]=sum ;
    }
  return 0 ;
}  

/* cholsl performs substitution of vector 'b' of size 'n' into Cholesky
decomposed (by 'choldc') matrix 'a' with diagonal vector 'p'.
On output the solution is returned in 'x' */

void cholsl(double *a,int n,double *p,double *b,double *x)
{ 
  int i,k ;
  double sum ;
  
  for(i=0;i<n;i++) {
    for(sum=b[i],k=i-1;k>=0;k--) sum -= a[i*n+k]*x[k] ;
    x[i]=sum/p[i] ;
  }
  for(i=n-1;i>=0;i--) {
    for(sum=x[i],k=i+1;k<n;k++) sum -= a[k*n+i]*x[k] ;
    x[i]=sum/p[i] ;
  }
}
#endif
