/* This is the implementation of the Kalman filter operating on a */
/* system consisting of two AR models, exactly as in darkalm3.m   */
/* Jan Kybic */


#ifndef MAKING_MEX
#include <stdlib.h>
#include <string.h>
#include "kalmse.h"

#define SAFEMALLOC(ptr,a,b) ptr=safemalloc((a)*(b)*sizeof(double)) ;
#define mxFree(a) safefree(a)
#define mexPrintf printf
#endif

#ifndef ABS
#define ABS(x) ((x)>0 ? (x) : -(x))
#endif

#define OVERFLOW_MULTIPLE (1.5)
/* if the output exceeds this time the maximum value, divergence is detected */

int darkalman3c(double *yp,double *zp,double *asp,
   double *bsp,double *csp,
   double *anp,double *bnp,double *cnp,int T,int ns, int nbs, int ncs, 
   int nn, int nbn, int ncn, double r
#ifndef MAKING_MEX 
   ,int unidir 
#endif
)
{ int i,j,k,n ; int flag=0 ;
  /* values */
  double qs,qn,e,tmp,tmp1,tmp2,Ls,maxz ;
  /* buffers */
  double *Pbp,*xbp,*Lbp,*ratbp,*ebp,*Almbp,*lmbp,*LLp,*ypb ;
  /* just pointers into buffers */
  double *Pp,*xp,*Lp,*ratp,*ep ;
  /* some other pointers */
  double *Prow ;

  n=ns+nn ; /* total order of the system */

  SAFEMALLOC(Pbp,T*n,n) 
  SAFEMALLOC(xbp,n,T)
  SAFEMALLOC(Lbp,n,T)
  SAFEMALLOC(ratbp,1,T)
  SAFEMALLOC(ebp,1,T) 
  SAFEMALLOC(Almbp,1,n)
  SAFEMALLOC(lmbp,1,n)
  SAFEMALLOC(LLp,1,n)

  qn=(*bnp)*(*bnp) ; /* noise subsystem input variance */
  qs=(*bsp)*(*bsp) ; /* signal subsystem input variance */

  /* set the pointers to the beginning*/
  Pp=Pbp ; xp=xbp ; Lp=Lbp ; ratp=ratbp ; ep=ebp ; ypb=yp ;

  /* initialize state */
  for(i=0;i<n;i++) xp[i]=0.0 ;

  /* initialize matrix P */
  for(i=0;i<n;i++)
    for(j=0;j<n;j++) Pp[i*n+j]=0.0 ;

  for(i=0,Prow=Pp ;i<nn;i++, Prow+=n) 
    { for(j=0;j<i;j++) Prow[j]=cnp[i-j] ;
      for(;j<nn;j++)   Prow[j]=cnp[j-i] ;
    }
  for(i=0,Prow+=nn;i<ns;i++, Prow+=n)
    { for(j=0;j<i;j++) Prow[j]=csp[i-j] ;
      for(;j<ns;j++)   Prow[j]=csp[j-i] ;
    }

  /* get the maximum input value for overflow control */
  for(maxz=0.0,k=0;k<T;k++) 
    if ((tmp=ABS(zp[k]))>maxz) maxz=tmp ;
  maxz*=OVERFLOW_MULTIPLE ;

  /* here begins the forward run */
  for(k=0;;k++) {
    /* data step */
    /* calculate and store prediction error */
    *ep=e=*zp++ - (xp[n-1]+xp[nn-1]) ; 
    *ratp=1/(Pp[(n-1)*n+(n-1)]+Pp[(nn-1)*n+nn-1]+2*Pp[(nn-1)*n+n-1]+r) ;

    /* calculate Kalman gain L and correct state estimate */
    tmp1=e*(*ratp) ;
    for(i=0;i<n;i++) { 
      LLp[i]=tmp=Pp[i*n+n-1]+Pp[i*n+nn-1] ; Lp[i]=*ratp * tmp ;
      xp[i]+=tmp*tmp1 ;
    }

    /* update covariance matrix */
    for(i=0;i<n;i++)
      for(j=0;j<n;j++) Pp[n*i+j]-=Lp[i]*LLp[j] ;

    if (k>=T-1) break ; /* do just data step for the last sample */

    /* time step */
    Pp+=n*n ; xp+=n ; Lp+=n ; ratp++ ; ep++ ; /* shift pointers */

    /* advance state vector */
    for(tmp=0.0,j=0;j<ns;j++) tmp+=asp[ns-j-1]*xp[-n+nn+j] ;
    for(j=0;j<ns-1;j++) xp[nn+j]=xp[nn+1-n+j] ;
    xp[n-1]=tmp ;
    for(tmp=0.0,j=0;j<nn;j++) tmp+=anp[nn-j-1]*xp[-n+j] ;
    for(j=0;j<nn-1;j++) xp[j]=xp[1-n+j] ;
    xp[nn-1]=tmp ;

    /* advance the covariance matrix as P=A*P*A'+Q */
    for(i=0;i<n;i++) { /* i is now the column index */
      for(tmp=0.0,j=0;j<ns;j++) tmp+=asp[ns-j-1]*Pp[-n*n+i+(nn+j)*n] ;
      for(j=0;j<ns-1;j++) Pp[i+(nn+j)*n]=Pp[-n*n+i+(nn+j+1)*n] ;
      Pp[i+(n-1)*n]=tmp ;
      for(tmp=0.0,j=0;j<nn;j++) tmp+=anp[nn-j-1]*Pp[-n*n+i+j*n] ;
      for(j=0;j<nn-1;j++) Pp[i+j*n]=Pp[-n*n+i+(j+1)*n] ;
      Pp[i+(nn-1)*n]=tmp ;
    }
    for(i=0;i<n;i++) { /* i is now the row index */
      for(tmp=0.0,j=0;j<ns;j++) tmp+=asp[ns-j-1]*Pp[i*n+(nn+j)] ;
      for(j=0;j<ns-1;j++) Pp[i*n+(nn+j)]=Pp[i*n+(nn+j+1)] ;
      Pp[i*n+(n-1)]=tmp ;
      for(tmp=0.0,j=0;j<nn;j++) tmp+=anp[nn-j-1]*Pp[i*n+j] ;
      for(j=0;j<nn-1;j++) Pp[i*n+j]=Pp[i*n+(j+1)] ;
      Pp[i*n+(nn-1)]=tmp ;
    }
    Pp[(n-1)*n+(n-1)]+=qs ;
    Pp[(nn-1)*n+(nn-1)]+=qn ;
  } 
  /* end of the forward run */

#ifndef MAKING_MEX
  if (unidir) goto finish ;
#endif

  /* here begins the backward run */  
  for(i=0;i<n;i++) lmbp[i]=0.0 ;  
  
  yp+=T-1 ;

  for(k=T-1;k>=0;k--) {
    Almbp[0]=lmbp[nn-1]*anp[nn-1] ;
    for(i=1;i<nn;i++) Almbp[i]=lmbp[i-1]+lmbp[nn-1]*anp[nn-i-1] ;
    Almbp[nn]=lmbp[n-1]*asp[ns-1] ;
    for(i=1;i<ns;i++) Almbp[i+nn]=lmbp[nn+i-1]+lmbp[n-1]*asp[ns-i-1] ;

    tmp1=0.0 ;
    for(j=0;j<n;j++) tmp1+=Pp[(n-1)*n+j] * Almbp[j] ;
    tmp2=0.0 ;
    for(j=0;j<n;j++) tmp2+=Pp[(nn-1)*n+j] * Almbp[j] ;

    *yp-- =(tmp=xp[n-1]+tmp1) ;
    if (ABS(tmp)>maxz) {
#ifdef MAKING_MEX
      mexPrintf(
       "DARKALM3C: Warning: Divergence detected. Backward run cancelled.\n") ;
#endif
      flag=-1 ;
      /* divergence detected - use only forward run */
    finish:
      for(i=0;i<T;i++) ypb[i]=xbp[n-1+i*n] ;
      break ;
    }

    Ls=Lp[n-1]+Lp[nn-1] ;
    tmp=(*ratp+Ls/r)*((1-Ls)*(*ep)-tmp1-tmp2) ;
    for(i=0;i<n;i++) lmbp[i]=Almbp[i] ;
    lmbp[nn-1]+=tmp ; lmbp[n-1]+=tmp ;
    
    Pp-=n*n ; xp-=n ; Lp-=n ; ratp-- ; ep-- ;
  }
  /* end of the backward run */

  mxFree(Pbp) ; mxFree(xbp) ; mxFree(Lbp) ; mxFree(ratbp) ; mxFree(ebp) ;
  mxFree(Almbp) ; mxFree(lmbp) ; mxFree(LLp) ; 
  return flag ;
}
