/* This is the implementation of the Kalman filter operating on a */
/* system consisting of two AR models, exactly as in darkalmsb.m   */
/* except that it does not calculate prediction error */
/* This version uses Cholesky decomposition */
/* It implements a Square Root Covariance Filter */
/* Jan Kybic */


#ifndef lint
static char vcid[] = "$Id$";
#endif /* lint */

#ifndef MAKING_MEX
#include <stdlib.h>
#include <string.h>

#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=malloc(a*b*sizeof(double)))==NULL) \
{ fprintf("DARKALM3C: could not malloc") ; exit(1) ; }
#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=malloc(a*b*sizeof(int)))==NULL) \
{ fprintf("DARKALM3C: could not malloc") ; exit(1) ; }
#define mxFree(a) free(a)
#define mexPrintf printf
#endif

#include "cholesk.c" /* the Cholesky decomposition routines */

#define ABS(x) ((x)>0 ? (x) : -(x))

#define SYMETRIZE 1 /* perform explicit symetrization ? (1=yes,0=no) */

#define LYAPTOL (1e-4)
#define LYAPMAXITER (300)
#define LYAPMINITER (10)
#define LYAPEPS (1e-20)

/* 'visualize' matrix 'x' with 'n' lines and 'm' columns */
static void visualize(double *x,int n,int m)
{ int i,j ;
  for(i=0;i<n;i++)
    { mexPrintf("\n") ;
      for(j=0;j<m;j++) 
	if (x[m*i+j]>0) 
	  mexPrintf("+") ;
	    else 
	      if (x[m*i+j]<0) mexPrintf("-") ;
	      else mexPrintf(".") ;
    }
  mexPrintf("\n") ;
}

/* 'triangularize' transforms matrix 'a' with 'n' lines and 'm'
columns to a triangular matrix 'y' sized m x m (m<n) by performing
chol(a'*a) If a'*a is singular, returns -1 as a failure indication, 0
otherwise. */
static int triangularize(double *a,int m,int n,double *y)
{ int i,j,k ;
  double sum ;

#if 0
  visualize(a,n,m) ;  
#endif

  for(i=0;i<m;i++)
    for(j=0;j<=i;j++) {
      sum=0.0 ;
      for(k=0;k<n;k++) sum+=a[i+m*k]*a[j+m*k] ;
      y[m*i+j]=y[m*j+i]=sum ;
    }

#if 0
  visualize(y,m,m) ;
#endif
  /* Now we have y=a'*a */
  return chol(y,m) ;
}
           

static  void initialize_P(double *P,int n,double *cnp,double *csp,int ns,int nn)
{ double *Prow ;
  int i,j ;
  /* initialize matrix P */
  for(i=0;i<n;i++)
    for(j=0;j<n;j++) P[i*n+j]=0.0 ;

  for(i=0,Prow=P ;i<nn;i++, Prow+=n) 
    { for(j=0;j<i;j++) Prow[j]=cnp[i-j] ;
      for(;j<nn;j++)   Prow[j]=cnp[j-i] ;
    }
  for(i=0,Prow+=nn;i<ns;i++, Prow+=n)
    { for(j=0;j<i;j++) Prow[j]=csp[i-j] ;
      for(;j<ns;j++)   Prow[j]=csp[j-i] ;
    }
}

static  void advance_state(double *x,double *xold,double *asp,double *anp,int ns, int nn, int n)
{ double tmp ;
  int j ;

  for(tmp=0.0,j=0;j<ns;j++) tmp+=asp[ns-j-1]*xold[nn+j] ;
  for(j=0;j<ns-1;j++) x[nn+j]=xold[nn+1+j] ;
  x[n-1]=tmp ;
  for(tmp=0.0,j=0;j<nn;j++) tmp+=anp[nn-j-1]*xold[j] ;
  for(j=0;j<nn-1;j++) x[j]=xold[1+j] ; 
  x[nn-1]=tmp ;
}


static  void right_mult(double *P1,double *P2,double *asp,
			      double *anp,int n, int nn, int ns)
/* P2=P1*A' */  
{ int i,j ;
  double tmp ;
  for(i=0;i<n;i++) { /* i is now the row index */
    for(tmp=0.0,j=0;j<ns;j++) tmp+=asp[ns-j-1]*P1[i*n+(nn+j)] ;
    for(j=0;j<ns-1;j++) P2[i*n+(nn+j)]=P1[i*n+(nn+j+1)] ;
    P2[i*n+(n-1)]=tmp ;
    for(tmp=0.0,j=0;j<nn;j++) tmp+=anp[nn-j-1]*P1[i*n+j] ;
    for(j=0;j<nn-1;j++) P2[i*n+j]=P1[i*n+(j+1)] ;
    P2[i*n+(nn-1)]=tmp ;
  }
}

static  void left_mult(double *P2,double *P1,double *asp,
			      double *anp,int n, int nn, int ns)
/* P1=A*P2 */
{ int i,j ; double tmp ;
    for(i=0;i<n;i++) { /* i is now the column index */
      for(tmp=0.0,j=0;j<ns;j++) tmp+=asp[ns-j-1]*P2[i+(nn+j)*n] ;
      for(j=0;j<ns-1;j++) P1[i+(nn+j)*n]=P2[i+(nn+j+1)*n] ;
      P1[i+(n-1)*n]=tmp ;
      for(tmp=0.0,j=0;j<nn;j++) tmp+=anp[nn-j-1]*P2[i+j*n] ;
      for(j=0;j<nn-1;j++) P1[i+j*n]=P2[i+(j+1)*n] ;
      P1[i+(nn-1)*n]=tmp ;
    }
}


/* 'dlyap' solves iteratively the Lyapunov equation A*X*A'+Q=X,
   given matrix A as 'asp','anp' initial estimate 'x' and constants
   'n','nn','qs','qn' */
static void  dlyap(double *asp,double *anp,double *x,int n,int nn,double qs,double qn)
{ int o,i,j,k,xsize=sizeof(double)*n*n,ns ;
  double *xold=alloca(xsize) ;
  double maxx,dif,tmp ;

  ns=n-nn ;

  for(o=0;o<LYAPMAXITER;o++) {
    memcpy(xold,x,xsize) ;
    right_mult(x,x,asp,anp,n,nn,ns) ;
    left_mult(x,x,asp,anp,n,nn,ns) ;
    /* now x=A*x*A' */
    x[(n-1)*n+(n-1)]+=qs ;
    x[(nn-1)*n+(nn-1)]+=qn ;

   if (o<LYAPMINITER) continue ;
   maxx=0.0 ; dif=0.0 ;
   for(i=0;i<n;i++)
     for(j=0;j<n;j++) {
       tmp=ABS(x[i*n+j]) ;
       if (tmp>maxx) maxx=tmp ;
       tmp=ABS(x[i*n+j]-xold[i*n+j]) ;
       if (tmp>dif) dif=tmp ;
     }
   if (dif/(maxx+LYAPEPS)<LYAPTOL) {
#if 0
     mexPrintf("%d iterations in dlyap\n",o) ;
#endif
     return ;
   }
  }
}

static void clear_matrix(double *x,int n,int m)
     /* sets matrix 'x' with 'n' lines and 'm' columns to zero */
{ int i,j ;
  for(i=0;i<n;i++)
    for(j=0;j<m;j++) *x++=0.0 ;
}

static void darkalmansc(double *yp,double *zp,double *asp,
   double *bsp,double *csp,
   double *anp,double *bnp,double *cnp,int T,int ns, int nbs, int ncs, 
   int nn, int nbn, int ncn, double r)
{ int i,j,k,n ;
  /* values */
  double qs,qn,e,tmp,tmp1,sr ;
  /* buffers */
  double *M,*Z1,*Z2,*Z4,*P1,*P2,*xfbp,*xpbp,*Lp,*LLp,*ypb,*Fb,*PTp,*Zp ;
  /* just pointers into buffers */
  double *xfp,*xpp,*Fp,*work,*zpb ;
  /* some other pointers */
  double *Pel ;
  int do_backward_run=1 ; /* a flag */

  n=ns+nn ; /* total order of the system */

  SAFEMALLOC(P1,n,n)
  SAFEMALLOC(P2,n,n) 
  SAFEMALLOC(xpbp,n,T+1)
  SAFEMALLOC(xfbp,n,T)
  SAFEMALLOC(Lp,1,n)
  SAFEMALLOC(LLp,1,n)
  SAFEMALLOC(Fb,n,T*n)
  SAFEMALLOC(PTp,n,n)
  SAFEMALLOC(work,n,1)
  SAFEMALLOC(M,n,n)
  SAFEMALLOC(Z1,2*n,n)
  SAFEMALLOC(Z2,n+1,n+1)
  SAFEMALLOC(Z4,n+1,n+1)


  qn=(*bnp) ; /* square of the noise subsystem input variance */
  qs=(*bsp) ; /* square of the signal subsystem input variance */
  sr=sqrt(r) ; /* square root of the measurement noise variance */

  /* set the pointers to the beginning*/
  xfp=xfbp ; xpp=xpbp ; /* filtered and predicted state estimates */  
  ypb=yp ; Fp=Fb ; zpb=zp ;

  /* initialize state */
  for(i=0;i<n;i++) xpp[i]=0.0 ;

  /* initialize matrix P */
  initialize_P(P1,n,cnp,csp,ns,nn) ;

  /* Solve Lyapunov equation */
  dlyap(asp,anp,P1,n,nn,qs,qn) ;
  
  /* Cholesky decomposition */
  memcpy(M,P1,n*n*sizeof(double)) ;
  if (chol(M,n)!=0) {
    mexPrintf(
"Initial covariance matrix non-positive definite in 'darkalmsc'\n") ;

  error:
    mexPrintf("  filtering skipped.\n") ;
    memcpy(ypb,zpb,T*sizeof(double)) ;
    return ;
  }

  /* initialize Z1,Z2 */
  clear_matrix(Z1,2*n,n) ;
  clear_matrix(Z2,n+1,n+1) ;
  Z2[0]=sr ;
  Z1[n*(n+nn-1)+nn-1]=qn ;
  Z1[n*(n+n-1)+n-1]=qs ;

  /* here begins the forward run */
  for(k=0;k<T;k++) {
    /* data step */
    /* calculate and store prediction error */
    e=*zp++ - (xpp[n-1]+xpp[nn-1]) ; 

    /* prepare Z2  */
    for(i=0;i<n;i++) {
      Z2[(n+1)*(i+1)]=M[i*n+n-1]+M[i*n+nn-1] ;
      for(j=0;j<n;j++)
	Z2[(i+1)*(n+1)+j+1]=M[i*n+j] ;
    }
    if (triangularize(Z2,n+1,n+1,Z4)!=0) { 
    mexPrintf(
"Failure to triangularize Z2 in 'darkalmsc'\n") ;

goto error ;
    /* the triangularized matrix is now in Z4 */
    }

    /* correct state estimate */
    tmp=1/Z4[0]*e ;
    for(i=0;i<n;i++) { 
      xfp[i]=xpp[i]+Z4[i+1]*tmp ;
    }

    /* update covariance matrix */
    for(i=0;i<n;i++)
      for(j=0;j<n;j++) M[n*i+j]=Z4[(i+1)*(n+1)+j+1] ;

    /* time step */
    xpp+=n ;  xfp+=n ; /* shift pointers */

    /* advance state vector */
    advance_state(xpp,xfp-n,asp,anp,ns,nn,n) ; 

    /* advance the covariance matrix */
    right_mult(M,Z1,asp,anp,n,nn,ns) ;
    /* now Z1=[M*A' ; Q ] */ 
    dechol(M,n,P2) ;
    right_mult(P2,P2,asp,anp,n,nn,ns) ;
    if (triangularize(Z1,n,2*n,M)!=0) {
    mexPrintf(
"Failure to triangularize Z1 in 'darkalmsc'\n") ;

    goto error ;
    }
    /* the new covariance factor is now in M */

    if (!do_backward_run) continue ;

    /* proceed to calculating F */    
    /* compute the full form and store it to P1 */
    dechol(M,n,P1) ;

    /* first of all, move P1 (transposed) to PTp */
    for(i=0;i<n;i++)
      for(j=0;j<n;j++) PTp[j*n+i]=P1[i*n+j] ;
   
    for(i=0;i<n;i++) {
      work[i]=M[i*n+i] ;
      for(j=i;j<n;j++) PTp[j*n+i]=M[i*n+j] ;
    }

    /* and backsubstitute */
    for(i=0;i<n;i++) cholsl(PTp,n,work,P2+i*n,Fp+i*n) ;

    Fp+=n*n ; /* advance the pointer */
  } 
  /* end of the forward run */

  if (do_backward_run) { 
    xfp-=n ; yp=ypb+(T-1) ; Fp-=n*n ;
    for(i=0;i<n;i++) work[i]=xpp[i] ;
    /* backward run */
    for(k=T-1;k>=0;k--) {
      for(i=0;i<n;i++) work[i]-=xpp[i] ;
      for(i=0;i<n;i++)
	for(j=0;j<n;j++) xfp[i]+=Fp[i*n+j]*work[j] ;
      for(i=0;i<n;i++) work[i]=xfp[i] ; 
    Fp-=n*n ; xfp-=n ; xpp-=n ;
    }
  /* end of the backward run */
  }

  
  for(i=0;i<T;i++) ypb[i]=xfbp[n-1+i*n] ;
      
  mxFree(P1) ; mxFree(P2) ; 
  mxFree(xpbp) ; mxFree(xfbp) ; mxFree(Lp) ; 
  mxFree(LLp) ; mxFree(Fb) ; mxFree(PTp) ; mxFree(work) ; 
  mxFree(Z1) ; mxFree(Z2) ; mxFree(M) ;
  mxFree(Z4) ;
  }
