/*
  darkalmbc    The MEX version of darkalm4 

  (except no prediction error is calculated)

% This function takes the corrupted signal z=signal+noise,
% AR parameters of signal (as,bs) and noise (an,bn)
% and signal resp. noise autocorrelations (cs,cn) and produces the smoothed
% estimate y of signal using bidirectional nonstationary Kalman filter
% r is the output measurement covariance. If not specified then 1e-6 * 
% input covariance is used
% The backward gain is *constant*

  The calling syntax is:

	y=darkalmbc(z,as,bs,cs,an,bn,cn,[r])

  Jan Kybic, October 1997

*/

#ifdef __GNUC__
#ifndef linux
#define linux
#endif
#endif

#ifdef linux
#define myCreateMatrix(a,b) mxCreateFull(a,b,REAL)
#define myAlloc(a) mxCalloc(a,1)
#else /* hp version */
#define myCreateMatrix(a,b) mxCreateDoubleMatrix(a,b,mxREAL)
#define myAlloc(a) mxMalloc(a)
#endif

#include <string.h>
#include "mex.h"

/* Input Arguments */

#define	Z_IN	prhs[0]
#define	AS_IN	prhs[1]
#define	BS_IN	prhs[2]
#define	CS_IN	prhs[3]
#define	AN_IN	prhs[4]
#define	BN_IN	prhs[5]
#define	CN_IN	prhs[6]
#define R_IN    prhs[7]

/* Output Arguments */

#define	Y_OUT	plhs[0]

/* Algorithm constants */
#define R_RATIO (1e-6)  /* relative measurement error squared, see code */

#define max(a,b) ((a>b) ? (a) : (b))

static void darkalmanbc(double *yp,double *zp,double *asp,
   double *bsp,double *csp,
   double *anp,double *bnp,double *cnp,int nz,int nas, int nbs, int ncs, 
   int nan, int nbn, int ncn,double r) ;

void printvector(double *v,int s)
{ int i ;
  mexPrintf("v=") ;
  for(i=0;i<s;i++) mexPrintf("%5g ",v[i]) ;
  mexPrintf("\n") ;
}

#ifdef linux
void mexFunction(
                 int nlhs, Matrix *plhs[],
                 int nrhs, Matrix *prhs[]
		 )
#else
void mexFunction(
                 int nlhs,       mxArray *plhs[],
                 int nrhs, const mxArray *prhs[]
		 )
#endif
{ int i,m,n,sizes[7] ;

  double *yp,*zp,*asp,*bsp,*csp,*anp,*bnp,*cnp,r ;
  int nz,ns,nbs,ncs,nn,nbn,ncn ;

  /* Check for proper number of arguments */
  
  if (nrhs != 7 && nrhs!=8) {
    mexErrMsgTxt("DARKALMBC requires seven or eight input arguments.");
  } else if (nlhs != 1) {
    mexErrMsgTxt("DARKALMBC requires one output argument.");
  }
  
  for (i=0;i<nrhs;i++)
    { m = mxGetM(prhs[i]);
      n = mxGetN(prhs[i]);
      sizes[i]=max(m,n) ;
      if (!mxIsNumeric(prhs[i]) || mxIsComplex(prhs[i]) || 
      mxIsSparse(prhs[i])  || !mxIsDouble(prhs[i]) || (m!=1 && n!=1)) 
         mexErrMsgTxt(
             "DARKALMBC needs all input arguments to be real vectors.");
     } 
  
  nz=sizes[0] ; ns=sizes[1] ;  nbs=sizes[2] ; ncs=sizes[3] ;
  nn=sizes[1] ;  nbn=sizes[2] ; ncn=sizes[3] ;

  Y_OUT = myCreateMatrix(1,nz);

  yp = mxGetPr(Y_OUT); zp=mxGetPr(Z_IN) ; 
  asp=mxGetPr(AS_IN) ; bsp=mxGetPr(BS_IN) ; csp=mxGetPr(CS_IN) ;
  anp=mxGetPr(AN_IN) ; bnp=mxGetPr(BN_IN) ; cnp=mxGetPr(CN_IN) ;
   
  if (nrhs==7)
    r=(*cnp+*csp)*R_RATIO ; /* measurement covariance */
  else
    r=*(mxGetPr(R_IN)) ;

  /*  %%%%% here the calculation begins %%%%%% */
  darkalmanbc(yp,zp,asp,bsp,csp,anp,bnp,cnp,nz,ns,nbs,ncs,nn,nbn,ncn,r) ;

  return;
}

#define SAFEMALLOC(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(double)))==NULL) \
    mexErrMsgTxt("DARKALMBC: could not malloc") ;

#define SAFEMALLOCINT(ptr,a,b) \
 if ((ptr=(void *)myAlloc((a)*(b)*sizeof(int)))==NULL) \
    mexErrMsgTxt("DARKALMBC: could not malloc") ;

#define MAKING_MEX
#include "../c/darkalm4.c"
