/* This is a source file for 'C' implemetations */
/* Noisy speech enhancement algorithm described by Gerhard Doblinger */
/* Jan Kybic */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "fft.h"
#include "dobl.h"

/* First we set the global parameters, see comments */
/* the current values are acceptable for 8000 samples/s and give 
   very good noise suppresion on the expense of some distortion */


#define WLEN 256     /* FFT window length */
#define POWER 8      /* 2^POWER=WLEN */
#define WSTEP 64     /* FFT window shift step size */
#define ALPHA 0.5    /* subband signal power smoothing factor */
                     /* set it to 0.5 - 0.7 */
#define BETA  0.96   /* nonlinear noise estimator constants */
#define GAMA  0.998  /* I do not know yet how to change them sensibly */

#define DELTA 15.0   /* noise bias factor - interesting to play with */
                     /* low values yield low noise suppresion
                        but low distortion. high values yield almost total 
                        noise elimination but unacceptable distortion 
                        Try from 2 to 50. I recommend 10-20 */

#define ETA   .95     /* weighting factor, not given in the article */
                      /* I do not think it has much influence */
                      
#define EPS   1e-20   /* small number, to avoid division by zero */ 

#define EULC  0.577215665 /* Euler constant C */

/* Instead of calculating the gain from the integral formula every single
   time, we store the values into a table. The following constants give
   the limits, table size and desired precision. There is not much point
   increasing the table size, moreover we do a linear interpolation */

#define MAXVKDB 15        /* max v_k in dB */
#define MINVKDB -30       /* min v_k in dB */
#define VKTABSIZE 1000    /* v_k table size */
#define VKPREC  1e-6      /* precision for v_k integral */


static readftype *getfp ;    /* pointer to the reading function */
static float inbuf[WLEN] ;   /* input buffer */
static complex spec[WLEN] ;  /* input spectrum */
static float px[WLEN] ;      /* smoothed spectral energy */
static float pxprev[WLEN] ;  /* previous smoothed spectral energy */
static float pn[WLEN] ;      /* noise power estimate */
static float g[WLEN] ;       /* gain */
static float accbuf[WLEN] ; /* accumulating buffer */
static complex tbuf[WLEN] ; /* here the frame output goes */
static float hann[WLEN] ; /* precomputed Hanning window coeficients */
static float pi ;         /* 3.1415... */
static char empty,first ; /* flags */
static float vktab[VKTABSIZE] ; /* table of g_k/r_k by v_k */


/* ********************************************************** */

/* reads num samples to dest using getfp 
   num must be multiple of chunk size WSTEP */ 
   
static int getfloat(float *dest,int num) 
{ int r ;  
  while(num>0)
   { /* read from the input */
     if ((r=(*getfp)(dest))!=0) return(r) ;
     /* and in(de)crement the counters */
     dest+=WSTEP ; num-=WSTEP ;
   }
  return(0) ;    
}   

/* ********************************************************** */

static void smooth()
{ /* spec -> px,pxprev */ 
  /* compute the current and smoothed spectral energy */
  
  float *fp,ax ; complex *cp ;
  
  memmove(pxprev,px,WLEN*sizeof(float)) ;
  
  for(cp=spec,fp=px;cp<&spec[WLEN];cp++,fp++)
    { ax=(cp->r)*(cp->r) + (cp->i)*(cp->i) ;
      *fp=*fp * ALPHA + ax * (1-ALPHA) ; 
    }
  
  if (first) memmove(pxprev,px,WLEN*sizeof(float)) ; 
}

/* ********************************************************** */


static void noise()
{ /* px,pxprev -> pn */ 
  /* This is the Doblinger's noise estimator based on some nonlinear formulas */  
     
  float *xptr,*xprevptr,*nptr ; 

  if (first) { memmove(pn,px,WLEN*sizeof(float)) ; return ; }

  for(xptr=px,nptr=pn,xprevptr=pxprev;xptr<px+WLEN;xptr++,xptr++,xprevptr++) 
       { if (*nptr>*xptr) *nptr=*xptr ;   
          else *nptr=GAMA*(*nptr)+(1-GAMA)/(1-BETA)*(*px-BETA*(*xprevptr)) ;
         /* Added 30.1.1998 - JK*/
         if (*nptr<0.0) *nptr=*xptr ;
       } 
         
}

/* ********************************************************** */


static void filter()
{ /* spec,px,pn -> conj(spec)/WLEN */
  /* Here we implement the filtering */
  
  float *xptr,*nptr,*gptr,*xprevptr ; complex *cp ;
  float post,prio,intgr,gain,rpost,r,v ;
  int fl ;
    
  /* go through the whole spectrum */
  for(cp=spec,xptr=px,nptr=pn,gptr=g,xprevptr=pxprev;cp<spec+WLEN;
        cp++,xptr++,nptr++,gptr++,xprevptr++)
   { 
     post=*xptr-DELTA*(*nptr) ;
     if (post<0) post=0 ;
     prio=(1-ETA)*post+ETA*(*gptr)*(*gptr)*(*xprevptr) ;
     rpost=post/(DELTA*(*nptr)+EPS) ;

     r=prio/(prio+DELTA*(*nptr)+EPS) ; 
     v=(1+rpost)*r ;

     if (v<1e-20) v=MINVKDB ;
             else v=10*log10(v) ;
     
     if (v>MAXVKDB) v=MAXVKDB ; else if (v<MINVKDB) v=MINVKDB ;
     v=(v-MINVKDB)*(VKTABSIZE-1)/(MAXVKDB-MINVKDB) ;
     fl=(int)floor(v) ;     
     if (fl>=VKTABSIZE-1) intgr=vktab[VKTABSIZE-1] ;
          else if (fl<=0) intgr=vktab[0] ;
          else intgr=vktab[fl]*(1-v+fl)+vktab[fl+1]*(v-fl) ;
     
     *gptr=gain=intgr*r ; 
 
     /* note we actually compute the complex cojugate of the
        output spectrum, divided by frame length to allow using
        FFT instead of IFFT */
     cp->r *= gain/WLEN ; cp->i *= -gain/WLEN ;            
    }  
}

/* ********************************************************** */

static void work()
{ 
  /* This routine takes one input frame and produces one
     output, hopefully noiseless output frame */

  float *fp,*ap ; complex *qp ;

  /* convert from real to complex and on the way apply the 
     Hanning window */
  for(ap=inbuf,qp=tbuf,fp=hann;ap<&inbuf[WLEN];ap++,fp++,qp++)
     { qp->r = *fp * (*ap) ; qp->i = 0 ; }
  
  /* compute the FFT of the input frame */
  fft(tbuf,spec) ;
  
  /* compute the amplitudes and the smoothed power spectrum */
  smooth() ;  

  /* run the noise estimator - reads px and generates pmin */ 
  noise() ;

  /* and finally we can apply the filtration 
     this gives us the new spectrum */
  filter() ; 
  
  /* note that filter inverts imaginary part of spec */
  /* so that we can compute IFFT using FFT simply as: */
  fft(spec,tbuf) ;
  
  /* note we discard imaginary part of tbuf so we do not have
     to bother with it any more and that filter() also makes
     the necesarry division */
 
  /* And that's it. The result stays in tbuf */
}

/* ********************************************************** */

int dobl_do()
{ /* This is the main function of the module. It is called whenever
     master wants next subframe. It takes care about assembling
     frames from subframes on input and overlapping windows on output */ 

  float *f ; complex *cp ;
  int r ;
  
  if (empty) /* input buffer empty ? */
    { empty=0 ; ;
      /* fill the input buffer */
      if ((r=getfloat(inbuf,WLEN))!=0) return(r) ;          
    }
   else /* input buffer is full */
    { /* discard the oldest part and  and move us forward */
      memmove(inbuf,inbuf+WSTEP,(WLEN-WSTEP)*sizeof(float)) ;
      /* and load the next chunk */
      if ((r=getfloat(inbuf+WLEN-WSTEP,WSTEP))!=0) return(r) ;
    }     
  
  /* move the accumulating buffer too */
  memmove(accbuf,accbuf+WSTEP,(WLEN-WSTEP)*sizeof(float)) ;
  /* zero the newly released part */
  for(f=accbuf+WLEN-WSTEP;f<&accbuf[WLEN];f++) *f=0.0 ;  
 
  /* now do the actual work ! */
  work() ;

  /* and add the results */
  for(f=accbuf,cp=tbuf;f<&accbuf[WLEN];) (*(f++)) += (cp++)->r ;
  
  first=0 ; /* reset the flag */
  /* And that's it. Wait till we are called again */
  return(0) ;
}  

/* *******************************************************  */

/* Integrate (exp(-x)-1)/x between 0 and uplim using Simpsons rule */

static double integral(double uplim,int div)
{ double step,sum,x ;
  int i ;

  step=uplim/div ; 
  sum=-1.0 ;                    /* x=0 */
  sum+=4*(exp(-step)-1)/step ;  /* x=step */
  x=2*step ;
  
  for(i=2;i<div;i+=2) 
    { sum+=2*(exp(-x)-1)/x ;
      x+=step ;
      sum+=4*(exp(-x)-1)/x ;
      x+=step ;
    }
  sum+=(exp(-uplim)-1)/x ; sum*=step/3 ;    
  
  return sum ;
}

/* Compute the integral above, honoring VKPREC */
static double integral_with_prec(double uplim)      
{ int div ;
  double sum,prevsum ;
  
  div=32 ; sum=integral(uplim,div) ;
  do {
    prevsum=sum ;
    div*=2 ; sum=integral(uplim,div) ;
  } while (fabs(sum-prevsum)>VKPREC) ;

  return sum ;
}



static void vktable_init() 
{
  int i ; double sum,vdb,x ;
  
  for(i=0;i<VKTABSIZE;i++)
   { vdb=MINVKDB+(double)i*(MAXVKDB-MINVKDB)/(VKTABSIZE-1) ;
     x=pow(10,vdb/10.0) ;
     sum=integral_with_prec(x) ;
     sum=-log(x)-EULC-sum ;
     vktab[i]=exp(0.5*sum) ;
   }  
}


/* ******************************************************** */ 
 
int dobl_init(readftype *f,int *sizep,float **outbuf,void *info)
 { 
   /* This function is called at the very beginning to establish
      the cooperation. Namely we find out the function to call
      when we need more data (f) and tell them our prefered
      chunk size (sizep) and where we place the output data
      (outbuf) */
   
   int r ;
   float *fp ; 

   /* store the getfp function pointer and set chunk size 
      and outbuf */
   getfp=f ; *sizep=WSTEP ; *outbuf=accbuf ;
   
   /* set the flags */
   empty=1 ; first=1 ; 
   
  /* set the accumulating output buffer to zero */
   for(fp=accbuf;fp<&accbuf[WLEN];fp++) *fp=0.0 ;       

  /* set the smooothed signal power to zero */
   for(fp=px;fp<&px[WLEN];fp++) *fp=0.0 ;       

  /* set the gain to one */
   for(fp=g;fp<&g[WLEN];fp++) *fp=1.0 ;       
  
   /* precompute Hanning coeficients */
   pi=acos(-1.0) ; /* Silly we have to do this.. */
   for(r=0;r<WLEN;r++) 
     hann[r]=((float)WSTEP/WLEN)*(1-cos(2*pi*r/(WLEN-1))) ;
   
   
   /* initialize FFT */
   if ((r=initfft(POWER))!=0) return(r) ;

   vktable_init() ;

   return(0) ;
 }           

/* ******************************************************** */ 
 
int dobl_close()
{ /* And this is to say goodbye. If we had any memory to 
     deallocate we would do it here. */ 

   closefft() ;
   return(0) ;
}  















