/* This is the main source file 'kalmse.c' for  
   Kalman filter based speech enhancement program 'kalmse'

   It is also used to build the kalmse_pvm variant to run under PVM

   Jan Kybic, October 1997
*/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <sys/times.h>
#include <unistd.h>
#include "kalmse.h"

/* global parameters of the algorithm */
kl_prms prms ;

static char *banner=
#ifndef KALMSEPVM
"This is kalmse, version " VERSION " by Jan Kybic\n" ;
#else
"This is kalmse_pvm, version " VERSION " by Jan Kybic\n" ;
#endif

static char *info=
#ifndef KALMSEPVM
"\nThis program (kalmse) is a speech enhancer for speech corrupted by additive\n"
#else
"\nThis program (kalmse_pvm) is a speech enhancer for speech corrupted by additive\n"
#endif
"   noise, based on Kalman filtering.\n"
"Usage: kalmse [options] input_file.bin output_file.bin\n"
"Options: (acceptable values [default value])\n"
"  -v  0-9  [9]             verbosity\n"
"  -nn 1-30 [8]             noise model order\n"
"  -ns 1-30 [8]             speech model order\n"
"  -wl 64,128,256,512 [256] window length\n"
"  -wo 1,2,4,8 [4]          window overlap factor\n"
"  -la 1-20 [8]             number of windows per subframe\n"
"  -lb 1-20 [8]             number of subframes\n"
"  -an 0.0-1.0 [0.9]        smoothing factor for noise estimate\n"
"  -as 0.0-1.0 [0.75]       smoothing factor for speech estimate\n"
"  -be 0.0-1.0 [0.9]        additional smoothing for noise estimate factor\n"
"  -eb 0.0-10.0 [1.5]       noise estimator bias\n"
"  -mc 0.0-10.0 [1e-6]      relative measurement error covariance\n"
"  -mh 0.0-10.0 [4]         maximum oversubtraction factor\n"
"  -ml 0.0-10.0 [1]         minimum oversubtraction factor\n"
"  -sr 0-4 [4]              subtraction rule (0=-,1=HWR,2=FWR,\n"
"                                              3=Wiener,4=Martin)\n"
"  -im 0-1 [0]              identification method (0=Burg,1=TSTLS)\n"
"  -tm 1-100000 [100]       max. number of iterations for TSTLS\n"
"  -kf 0-2 [1]              Kalman filter (0=unidir,1=bidir,2=always SR)\n" 
"  -fm 0-1 [0]              Spectrum estimation (0=MEM,1=FFT)\n"
#ifdef KALMSEPVM
"  -pv 1-1000 [1]           How many slaves to spawn\n"
"  -sl 0-1    [0]           Run as slave (for internal use only!)\n"
#endif                 
"We assume 16bit, signed, mono, 8000Hz signal with no header.\n"
"For more information, please refer to the documentation.\n" ;

static FILE *infile,*outfile ; /* input and output files */


/* ----------------------------------------------------------------------- */

/* set parameters to defaults, for description see kalmse.h */
void default_parameters(kl_prms *p) 
{ p->v=9 ;
  p->nn=8 ;
  p->ns=8 ;
  p->wl=256 ;
  p->wo=4 ;
  p->la=8 ;
  p->lb=8 ;
  p->alpha=0.9 ;
  p->alphas=0.75 ;
  p->be=0.9 ;
  p->eb=1.5 ;
  p->mc=1e-6 ;
  p->mh=4.0 ;
  p->ml=1.0 ;
  p->sr=4 ;
  p->im=0 ;
  p->tsmax=100 ;
  p->kf=1 ;
  p->fm=0 ;
  p->inbuf=NULL ; p->outbuf=NULL ;
#ifdef KALMSEPVM
  p->pv=1 ; p->sl=0 ;
#endif
}

/* ----------------------------------------------------------------------- */

void rprintf(int verb,char *f,...) 
{ 
  va_list ap;
  
  if (verb>prms.v) return  ;
  
  va_start (ap, f);
  vprintf (f, ap);
  va_end (ap);
  fflush(stdout) ;
}
/* ----------------------------------------------------------------------- */

void errp(char *f,...) 
{ 
  va_list ap;

  printf("kalmse: ") ;
  va_start (ap, f);
  vprintf (f, ap);
  va_end (ap);
  printf(" %s\n",strerror(errno)) ;
  fflush(stdout) ;
}

/* ----------------------------------------------------------------------- */

void eprintf(int verb,char *f,...) 
{ 
  va_list ap;
  
  if (verb>prms.v) return  ;
  
  printf("kalmse: ") ;
  va_start (ap, f);
  vprintf (f, ap);
  va_end (ap);
  fflush(stdout) ;
}

/* ----------------------------------------------------------------------- */

void show_parameters(kl_prms *p) 
{
  rprintf(PROCL,"\nParameters in use:\n") ;
  rprintf(PROCL,
	  "Verbosity: %d. Speech model order: %d. Noise model order: %d.\n",
	  p->v,p->ns,p->nn) ;
  rprintf(PROCL,"Window length: %d. Window overlap: %d.\n",p->wl,p->wo) ;
  rprintf(PROCL,"Windows per subframe: %d. Subframes: %d.\n",p->la,p->lb) ;
  rprintf(PROCL,"Speech smoothing factor: %g Noise smoothing factor: %g\n",
	  p->alpha,p->alpha) ;
  rprintf(PROCL,"Additional noise estimate smoothing factor: %g\n",p->be) ;
  rprintf(PROCL,"Noise estimator bias: %g\n",p->eb) ;
  rprintf(PROCL,"Relative measurement error covariance: %g\n",p->mc) ;  
  rprintf(PROCL,"Minimum/maximum oversubtraction factor: %g/%g\n",
	  p->mh,p->ml) ;
  rprintf(PROCL,"Subtraction rule: %d. Identification method: %d.\n",
	  p->sr,p->im) ;
  rprintf(PROCL,"Maximum number of iterations for TSTLS: %d.\n",p->tsmax) ;
  rprintf(PROCL,"Kalman filter method: %d\n",p->kf) ;
  rprintf(PROCL,"Spectrum estimation  method: %d\n",p->fm) ;
#ifdef KALMSEPVM
  rprintf(PROCL,"Number of slaves: %d. Slave flag: %d.\n",p->pv,p->sl) ;
#endif
  rprintf(PROCL,"\n") ;
}

/* ----------------------------------------------------------------------- */

#define INTOPT(name,minv,maxv,where) \
len=strlen(name) ; \
if (strncmp(s,name,len)==0) \
{ if (*(s+len)) iv=atoi(s+len) ; else iv=atoi(argv[++i]) ; \
  if (iv<minv || iv>maxv) { \
    rprintf(INFOL,banner) ; \
    eprintf(ERRL,"Option '%s' out of range.\n",name) ; \
    return -1 ; } \
  p->where=iv ; \
  continue ; }

#define FLTOPT(name,minv,maxv,where) \
len=strlen(name) ; \
if (strncmp(s,name,len)==0) \
{ if (*(s+len)) fv=atof(s+len) ; else fv=atof(argv[++i]) ; \
  if (fv<minv || fv>maxv) { \
    rprintf(INFOL,banner) ; \
    eprintf(ERRL,"Option '%s' out of range.\n",name) ; \
    return -1 ; } \
  p->where=fv ; \
  continue ; }

int get_parameters(kl_prms *p,int argc,char *argv[]) 
{ int i ; 
  char *s ; 
  int iv,len ; 
  float fv ;
  
/* go through the arguments, skipping the program name */
 for (i=1;i<argc;i++) { 
   if (*argv[i]!='-') break ; 
   s=argv[i]+1 ;
   INTOPT("v",0,9,v) 
   INTOPT("nn",1,30,nn)    
   INTOPT("ns",1,30,ns)
   INTOPT("wl",64,512,wl)
   if (p->wl!=64 && p->wl!=128 && p->wl!=256 && p->wl!=512) {
     rprintf(INFOL,banner) ;
     eprintf(ERRL,"Invalid value for option 'wl'.\n") ;
     return -1 ;
   }
   INTOPT("wo",1,8,wo)
   if (p->wo!=1 && p->wo!=2 && p->wo!=4 && p->wo!=8) {
     rprintf(INFOL,banner) ;
     eprintf(ERRL,"Invalid value for option 'wo'.\n") ;
     return -1 ;
   }
   INTOPT("la",1,20,la)
   INTOPT("lb",1,20,lb)
   FLTOPT("an",0.0,1.0,alpha)
   FLTOPT("as",0.0,1.0,alphas)
   FLTOPT("be",0.0,1.0,be)
   FLTOPT("eb",0.0,10.0,eb)
   FLTOPT("mc",0.0,10.0,mc)
   FLTOPT("mh",0.0,10.0,mh)
   FLTOPT("ml",0.0,10.0,ml)
   INTOPT("sr",0,4,sr)
   INTOPT("im",0,1,im)
   INTOPT("tm",1,100000,tsmax)
   INTOPT("kf",0,2,kf)   
   INTOPT("fm",0,1,fm)
#ifdef KALMSEPVM
   INTOPT("pv",1,1000,pv)
   INTOPT("sl",0,1,sl)
#endif
   rprintf(INFOL,banner) ;
   eprintf(ERRL,"Unknown option '%s'\n",argv[i]) ;
   return -1 ;
   }
return i ; 
}

/* ----------------------------------------------------------------------- */
  
int main(int argc,char *argv[])
{ int i ;

  if (sizeof(WORD)!=2) {
    eprintf(ERRL,"Incorrect definition of WORD in 'kalmse.h'.\n") ;
    return -1 ;
  }

  default_parameters(&prms) ;
#ifndef KALMSEPVM

  if (argc<3) {
    rprintf(INFOL,banner) ;
    rprintf(INFOL,info) ;
    return -1 ;
  }
  if ((i=get_parameters(&prms,argc,argv))<0) return -1 ;

  rprintf(INFOL,banner) ;

  if (i+2>argc) {
    eprintf(ERRL,"Two file names expected.\n") ;
    return -1 ; }

  if ((infile=fopen(argv[i],"r"))==NULL) {
    errp("Could not open file '%s' for reading.",argv[i]) ;
    return -1 ; }

  if ((outfile=fopen(argv[i+1],"w"))==NULL) {
    fclose(infile) ;
    errp("Could not open file '%s' for writing.",argv[i+1]) ;
    return -1 ; }

  show_parameters(&prms) ; 

  filter_file(&prms,infile,outfile) ;

  fclose(infile) ; fclose(outfile) ;
#else
  if ((i=get_parameters(&prms,argc,argv))<0) return -1 ;
  rprintf(INFOL,banner) ;
  if (prms.sl==0) {
    /* I am the master process */
    rprintf(INFOL,banner) ;
    if (argc<3) {
      rprintf(INFOL,info) ;
      return -1 ;
    }
  
    if (i+2>argc) {
      eprintf(ERRL,"Two file names expected.\n") ;
      return -1 ; }

    if ((infile=fopen(argv[i],"r"))==NULL) {
      errp("Could not open file '%s' for reading.",argv[i]) ;
      return -1 ; }

    if ((outfile=fopen(argv[i+1],"w"))==NULL) {
      fclose(infile) ;
      errp("Could not open file '%s' for writing.",argv[i+1]) ;
      return -1 ; }

    show_parameters(&prms) ; 

    filter_file(&prms,infile,outfile) ;

    fclose(infile) ; fclose(outfile) ;
  }
  else be_slave(&prms) ;
#endif  

  return 0 ;
}

/* ----------------------------------------------------------------------- */
void *safemalloc(int size)
{ void *ptr ;

  if ((ptr=malloc(size))==NULL) {
    eprintf(ERRL,"Not enough memory.") ;
    exit(-1) ;
  }
  return ptr ;
}

/* ----------------------------------------------------------------------- */

void safefree(void *ptr) { free(ptr) ; }

/* ----------------------------------------------------------------------- */

void clean_up_buffer(kl_prms *p) 
{ 
 safefree(p->inbuf) ; safefree(p->outbuf) ;
}

/* ----------------------------------------------------------------------- */

void read_bin_file(kl_prms *p,FILE *infile) 
{ int i ; double v ;
  WORD *ibuf ;

 if (fseek(infile,0L,SEEK_END)) { 
   errp("Could not fseek on input file.") ;
   exit(-1) ;
 }
 p->T=(ftell(infile)+1)/2 ;
 p->wshft=p->wl / p->wo ;
 /* round buflen up to the nearest multiple of wshft */
 p->buflen=((p->T + p->wshft - 1)/p->wshft) * p->wshft ;
 /* assure minimal length */
 if (p->buflen<p->wl) p->buflen=p->wl ;
 p->nwin=p->buflen / p->wshft - p->wo + 1 ; 
 rprintf(PROCL,"Input file contains %d samples, i.e. %d windows\n",
	 p->T,p->nwin) ;

 rewind(infile) ; 
 
 p->inbuf=safemalloc(sizeof(double) * p->buflen) ;
 p->outbuf=safemalloc(sizeof(double) * p->buflen) ; 
 ibuf=safemalloc(sizeof(WORD) * p->buflen) ;    

 if (p->T!=fread(ibuf,sizeof(WORD),p->T,infile)) {
   eprintf(ERRL,"Failed to read from the input file.\n") ;
   exit(-1) ;
 }

 rprintf(DBGL,"Input file sucessfully read.\n") ;

 p->mean=0.0 ;
 for(i=0;i<p->T;i++) p->mean+=(double)ibuf[i] ;
 p->mean/=p->T ;

 p->ampl=0.0 ;
 for(i=0;i<p->T;i++) 
   if ((v=fabs((double)ibuf[i]-p->mean))>p->ampl) p->ampl=v ;

 rprintf(DBGL,"Input data mean=%g, amplitude=%g\n",p->mean,p->ampl) ;

 for(i=0;i<p->T;i++)
   p->inbuf[i]=((double)ibuf[i] - p->mean)/p->ampl ;

 safefree(ibuf) ;  
}

/* ----------------------------------------------------------------------- */

void write_bin_file(kl_prms *p,FILE *outfile)
{ int i  ; double v ;
  WORD *ibuf ;

  ibuf=safemalloc(sizeof(WORD) * p->T) ;    

  for(i=0;i<p->T;i++) { 
    v=rint(p->outbuf[i] * p->ampl + p->mean) ;
    if (v>MAXWORD) v=MAXWORD ; else 
      if (v<MINWORD) v=MINWORD ;
    ibuf[i]=(WORD)v ;
  }

  if (p->T!=fwrite(ibuf,sizeof(WORD),p->T,outfile)) {
   eprintf(ERRL,"Failed to write to the output file.\n") ;
   exit(-1) ;
  }

  rprintf(DBGL,"Output file sucessfully written.\n") ;
  safefree(ibuf) ;
}
/* ----------------------------------------------------------------------- */


void filter_file(kl_prms *p,FILE *infile,FILE *outfile) 
{ struct tms t1,t2 ; clock_t t1r,t2r ;

 int CLK_TCK = sysconf(_SC_CLK_TCK) ;

  read_bin_file(p,infile) ;
  rprintf(INFOL,"Filtering started.\n") ;
  
  t1r=times(&t1) ;
#ifndef KALMSEPVM
  filter_buffer(p) ;
#else
  filter_buffer_master(p) ;
#endif
  t2r=times(&t2) ;

  rprintf(INFOL,"\nFiltering used %3fs (real) %3fs (user) %3fs (sys).\n",
	  ((double)t2r-t1r)/CLK_TCK,
	  ((double)t2.tms_utime-t1.tms_utime)/CLK_TCK,
	  ((double)t2.tms_stime-t1.tms_stime)/CLK_TCK) ;

  write_bin_file(p,outfile) ;
  clean_up_buffer(p) ; 
}

/* ---------------------- end of kalmse.c ---------------------------------- */


