/* This is a header file 'kalmse.h' for 
   Kalman filter based speech enhancement program 'kalmse'

   Jan Kybic, October 1997
*/

#ifndef _kalmse_h
#define _kalmse_h

#include <stdio.h>
#include "fft.h"

#define VERSION "1.2 (8.1.1998)"

/* the verbosity levels */
#define ERRL  1
#define WARNL 2
#define INFOL 3
#define PROCL 5
#define DBGL  9

#ifdef KALMSEPVM 
#include "pvm3.h"
#define NPENDWIND 3   /* number of pending windows for each slave */
#define SLAVENAME "kalmse_pvm"
#define SLAVEARGS "-sl1","-v1"
  struct slinfo { /* slave info */
    int ws ; } ;
#endif

typedef struct kl_prms {
  /* this structure contains all the information needed for the algorithm */
  /* first the user changeable parameters */
  int v ;  /* verbosity level */
  int nn ; /* noise model order */
  int ns ; /* speech model order */
  int wl ; /* window length */
  int wo ; /* window overlap */
  int la ; /* number of windows per subframe for noise estimation */
  int lb ; /* number of subframes for noise estimation */
  float alpha ; /* smoothing factor for noise */
  float alphas ; /* smoothing factor for speech */
  float be ; /* additional smoothing of noise estimate */
  float eb ; /* noise estimate bias factor */
  float mc ; /* measurement error covariance */
  float mh ; /* maximum oversubtraction */
  float ml ; /* minimum oversubtraction */
  int sr ; /* subtraction rule (0=simple,1=HFR,2=TWR,3=Wiener,4=Martin) */
  int im ; /* identification method (0=burg,1=TSTLS) */
  int kf ; /* Kalman filter type (0=unidirectional,1=fast bidir,2=slow SR) */
  int fm ; /* Frequency estimation method (0=MEM,1=FFT) */
  int tsmax ; /* maximum number of iterations for TSTLS */
#ifdef KALMSEPVM 
  int pv,sl ; /* number of slaves. am I a slave? */
#endif

  /* now the internal data */
  double *inbuf,*outbuf ; /* input and output buffers */
  int buflen ;            /* buffer length */
  int T ;                 /* number of samples */
  int nwin ;              /* number of windows */ 
  int wshft ;             /* window shift */ 
  double mean,ampl ;      /* mean and amplitude of the input signal */
  double *hann ;          /* Hanning window coefficients */
  int wllog ;             /* binary logarithm of window length */ 
  complex *cbuf1,*cbuf2 ;  /* complex buffers for FFT */
  double *sps,*ssps ;     /* smoothed power spectra */
  double *circb,*cbmin ;  /* circular buffer and its minimum */
  double *lsmin ;         /* minimum of the last subframes */
  double *nps ;           /* noise power spectrum estimate */
  double *snps ;          /* smoothed noise power spectrum estimate */
  double *buf ;           /* a buffer */ 
  int cbi ;               /* circular buffer index */
  int lsc ;               /* last segment count */
  double *sss ;           /* speech spectrum estimate */
  double *cs,*cn ;        /* autocorrelations */
  double *an,bn,*as,bs ;  /* AR model parameters */
#ifdef KALMSEPVM 
  int *tids ;             /* TIDs of slaves */
  struct slinfo *si ;     /* slave info */
  int pend ;              /* number of pending tasks */
#endif

} kl_prms ;

extern kl_prms prms ;

/* we need WORD to be two bytes */
#define WORD short int
#define MAXWORD (32767)
#define MINWORD (-32768)

#define MIN(a,b) ((a)>(b) ? (b) : (a))
#define MAX(a,b) ((a)>(b) ? (a) : (b))
#define ABS(a) ((a)>0 ? (a) : (-(a)))

#define EPS (1e-30) /* a very small number */
#define MAXOVERSHOOT (1.0) /* if the relative amplitude of the
  filtered output is greater than MAXOVERSHOOT, divergence is assumed
  and the SR implementation is used */


/* set parameters to defaults */
void default_parameters(kl_prms *p) ; 

/* rprintf prints a message if the current verbosity level is higher 
   or equal to verb */
void rprintf(int verb,char *f,...) ;

/* eprintf is like rprintf but precedes the message by a program name */
void eprintf(int verb,char *f,...) ;

/* errp prints an error message together with error given by the system 
   The error message should not be ended by \n */
void errp(char *f,...) ; 

/* get_parameters parses the command line and sets the parameters
   returns the next command line item index (first file name) or -1 if
   error occured */
int get_parameters(kl_prms *p,int argc,char *argv[]) ;

/* show_parameters uses rprintf to display the parameters in use */
void show_parameters(kl_prms *p) ;

/* filter_file takes opened infile, performs filtering and writes the results
   to outfile */
void filter_file(kl_prms *p,FILE *infile,FILE *outfile) ;

/* safemalloc - like malloc but exits if there is not enough memory */
void *safemalloc(int size) ;

/* safefree - should be used with safemalloc, currently equivalent to free */
void safefree(void *ptr) ;
 
/* read_bin_file reads the binary infile into the buffer, allocating the memory and setting its length and number of frames */
void read_bin_file(kl_prms *p,FILE *infile) ;

/* filter buffer performs the actual filtering */
void filter_buffer(kl_prms *p) ;

/* write_bin_file writes the output buffer contents to the binary outfile */
void write_bin_file(kl_prms *p,FILE *outfile) ;

/* clean_up_buffer deallocates the buffers */
void clean_up_buffer(kl_prms *p) ; 

/* prepare the filtering */
void filter_init(kl_prms *p) ;

/* clean up after the filtering */
void filter_close(kl_prms *p) ;

/* a helper functions for filter buffer */
void calculate_power_spectrum(kl_prms *p,int ind) ;
void smooth_power_spectra(kl_prms *p) ;
void estimate_noise(kl_prms *p) ;
void subtraction_rule(kl_prms *p) ;
void kalman_filter(kl_prms *p,int ind) ;
void assemble_window(kl_prms *p,int ind) ;

/* estimate AR parameters given a spectrum */
void estimate_AR_parameters(kl_prms *p,int n,double *a,
			    double *b, double *sp, double *c) ;

/* Burg algorithm */
void burg( double *x, double *a, int n, int T) ;

/* Total Structured Toeplitz Least Squares algorithm using wieghting */
void tstlsw(double *xp,double *ap,int n,int T,int maxiter) ;

/* Fast bidirectional Kalman filtering */ 
int darkalman3c(double *yp,double *zp,double *asp,
   double *bsp,double *csp,
   double *anp,double *bnp,double *cnp,int nz,int nas, int nbs, int ncs, 
   int nan, int nbn, int ncn,double r,int unidir) ;

/* Slow, square root covariance implementation of Kalman filtering */
void darkalmansc(double *yp,double *zp,double *asp,
   double *bsp,double *csp,
   double *anp,double *bnp,double *cnp,int T,int ns, int nbs, int ncs, 
   int nn, int nbn, int ncn, double r) ;

#ifdef KALMSEPVM
void be_slave(kl_prms *p) ; /* perform all the hard work */
void filter_buffer_master(kl_prms *p) ; /* parallel version of filter_buffer */
int process_message(kl_prms *p) ; /* receive and process message, returns the 
				     slave number  */
void give_work(kl_prms *p,int i,int sn) ; /* send window i to slave j */
#endif



#endif

/* ---------------------- end of kalmse.h ---------------------------------- */

