/* This file contains routines for LU decomposition and backsubstitution
   used for solving linear equation sets. The routines are based heavily
   on those from 'Numerical Recipes in C, Second Edition, Cambridge 
   University Press'. I typed the originals into the computer myself, using 
   legally obtained book, which is, as far as I understand, permitted by the
   'immediate license'. However, this license does not authorize you to use 
   those routines for whatever purposes, unless you get the book and type them
   in yourself or get the appropriate license from 'Numerical Recipes Software'
   
   I might eventually rewrite these routines or use some freely available 
   library like LAPACK instead but for the time being, you just have to put
   up with the above mentioned limitations.

   Jan Kybic, October 1997
*/

/* this file can be compiled separately or '#include'd (for use in MEX files)*/


#ifndef _LUDECOMP_C
#define _LUDECOMP_C
#include <math.h>
#define TINY 1e-20

/* ludcmp takes a matrix a, size n*n and LU decomposes it in place.
   'vv' should point to a work place of size n*sizeof(double)
   On output 'indx' contains the row permutation. Returns zero on
   successful return or -1 if the matrix was singular */

int ludcmp(double *a,int n, int *indx,double *vv)
{
  int i,imax,j,k ;
  double big,dum,sum,temp ;
  /* 'vv' stores implicit scaling of each row. We could have used
     'malloc' to allocate it in place - which is slow, or 'alloca' -
     which is not portable, so we decided to pass the responsibility
     on the caller */

  for(i=0;i<n;i++) {
    big=0.0 ;
    for(j=0;j<n;j++)
      if ((temp=fabs(a[i*n+j]))>big) big=temp ;
    if (big==0.0) return -2 ; /* singular matrix */
    vv[i]=1.0/big ;
  }

  for(j=0;j<n;j++) {
    for(i=0;i<j;i++) {
      sum=a[i*n+j] ;
      for(k=0;k<i;k++) sum-=a[i*n+k]*a[k*n+j] ;
      a[i*n+j]=sum ;
    }
    big=0.0 ;
    for(i=j;i<n;i++) {
      sum=a[i*n+j] ;
      for(k=0;k<j;k++) sum-=a[i*n+k]*a[k*n+j] ;
      a[i*n+j]=sum ;
      if ((dum=vv[i]*fabs(sum))>=big) {
	/* is the figure of merit best so far ? */
	big=dum ; imax=i ;
      }
    }
    if (j!=imax) { /* interchange rows if needed be */
      for(k=0;k<n;k++) {
	dum=a[imax*n+k] ;
	a[imax*n+k]=a[j*n+k] ;
	a[j*n+k]=dum ;
      }
      vv[imax]=vv[j] ;
    }
    indx[j]=imax ;
    if (a[j*n+j]==0.0) a[j*n+j]=TINY ;
    if (j != n-1) {
      dum=1.0/(a[j*n+j]) ;
      for (i=j+1;i<n;i++) a[i*n+j]*=dum ;
    }
  }
  return 0;
}


/* lubksb performs substitution of vector 'b' of size 'n' into LU
decomposed (by 'ludcpm') matrix 'a' with permutation vector 'indx'.
On output 'b' is overwritten with the solution */

void lubksb(double *a,int n, int *indx,double *b)
{ 
  int i,ii=-1,ip,j ;
  double sum ;
  
  for(i=0;i<n;i++) {
    ip=indx[i] ; sum=b[ip] ; b[ip]=b[i] ;
    if (ii>-1)
      for(j=ii;j<i;j++) sum-=a[i*n+j]*b[j] ;
    else if (sum!=0.0) ii=i ;
    b[i]=sum ;
  }

  for (i=n-1;i>=0;i--) {
    sum=b[i] ;
    for(j=i+1;j<n;j++) sum-=a[i*n+j]*b[j] ;
    b[i]=sum/a[i*n+i] ;
  }
}

#endif
