/* This is a source file for 'C' implemetations */
/* Noisy speech enhancement algorithm described by Reiner Martin */
/* Jan Kybic */


#include <string.h>
#include <math.h>
#include "fft.h"
#include "martin.h"

/* First we set the global parameters */

#define WLEN 256     /* FFT window length */
#define POWER 8      /* 2^POWER=WLEN */
#define WSTEP 64     /* FFT window shift step size */
#define ALPHA 0.9    /* subband signal power smoothing factor */
#define SUBF  0.02   /* spectral floor constant */
#define OMIN  1.5    /* correction factor to compensate bias of noise estimate */
#define M     8      /* minimum estimate window length */
#define W     7      /* number of minimum estimate windows */     

static readftype *getfp ;    /* pointer to the reading function */
static float inbuf[WLEN] ;   /* input buffer */
static complex spec[WLEN] ;  /* input spectrum */
static float ax[WLEN] ;      /* current spectral energy */
static float px[WLEN] ;      /* smoothed spectral energy */
static float pmact[WLEN] ;   /* current noise minimum */
static float pmin[WLEN] ;    /* noise power minimum */
static float pmmin[W][WLEN] ; /* noise minimum in last W M-subwindows */
static float accbuf[WLEN] ; /* accumulating buffer */
static complex tbuf[WLEN] ; /* here the frame output goes */
static float hann[WLEN] ; /* precomputed Hanning window coeficients */
static float pi ;         /* 3.1415... */
static char empty,first ; /* flags */
static int mi ;           /* number of subwindows read */


/* ********************************************************** */

/* reads num samples to dest using getfp 
   num must be multiple of chunk size WSTEP */ 
   
static int getfloat(float *dest,int num) 
{ int r ;  
  while(num>0)
   { /* read from the input */
     if ((r=(*getfp)(dest))!=0) return(r) ;
     /* and in(de)crement the counters */
     dest+=WSTEP ; num-=WSTEP ;
   }
  return(0) ;    
}   

/* ********************************************************** */

static void smooth()
{ /* spec -> ax, px */ 
  /* compute the current and smoothed spectral energy */
  
  float *fp,*ap ; complex *cp ;
  
  for(cp=spec,ap=ax,fp=px;cp<&spec[WLEN];cp++,ap++,fp++)
    { *ap=(cp->r)*(cp->r) + (cp->i)*(cp->i) ;
      *fp=*fp * ALPHA + *ap * (1-ALPHA) ; 
    }
}

/* ********************************************************** */


static void noise()
{ /* px -> pmin */ 
  /* This is the noise estimator based on minimum tracking */  
  /* We group the frames into subwindows of length M. W of these
     form a window. (I admit the terminology is confusing.) */
     
  int k ; float s ;
  float *fp,*ap,*bp ; 

  /* In the first subwindow frame, store current energy spectrum.
     For the following ones, track the minimum */ 
  if (mi==0) memmove(pmact,px,WLEN*sizeof(float)) ;
     else   for(fp=pmact,ap=px;fp<&pmact[WLEN];fp++,ap++) 
               if (*ap<*fp) *fp=*ap ;   
    
  /* at the end of each subwindow, store the minimum and start again fresh */  
  if (++mi > M) /* M subwindows passed */
   { mi=0 ;
     if (first) /* the very first time we copy the spectrum everywhere */
      { first=0 ;
        for(k=0;k<W;k++) memmove(pmmin[k],px,WLEN*sizeof(float)) ;
      }
 else {  
        /* the other times, shift the table left and store the new
           spectrum into the rightmost column */         
        memmove(pmmin[0],pmmin[1],(W-1)*WLEN*sizeof(float)) ; 
        memmove(pmmin[W-1],pmact,WLEN*sizeof(float)) ;
        
        /* compute pmin as column-wise minimum of pmmin */
        for(fp=pmin,ap=(float *)pmmin;fp<&pmin[WLEN];fp++,ap++)
          { s=*ap ;
            for(bp=ap+WLEN;bp<&pmmin[W][0];bp+=WLEN)
              if (*bp<s) s=*bp ;
            *fp=s ;
          }            
      }
    }  

  /* and then set -during the first M frames- or update pmin */
  if (first) memmove(pmin,px,WLEN*sizeof(float)) ;
     else   for(fp=pmin,ap=px;fp<&pmin[WLEN];fp++,ap++) 
               if (*ap<*fp) *fp=*ap ;  

}

/* ********************************************************** */


static void filter()
{ /* px,ax,spec,pmin -> conj(spec)/WLEN */
  /* Here we implement the filtering */
  
  float *fp,*ap,*bp ; complex *cp ;
  float s,pn ;
  
  /* go through the whole spectrum */
  for(cp=spec,fp=pmin,bp=px,ap=ax;cp<&spec[WLEN];cp++,fp++,ap++,bp++)
   { 
     /* pn is the noise estimate */
     pn=OMIN* (*fp) ;    

     /* we use s as a temporary storage place here */
     /* the if is here to avoid dividing by zero */      
     if (pn<=*bp)
       { s=(*bp-pn)/pn ; /* compute the normalized difference */
         /* limit it to the range 1 .. 100 */ 
         if (s<1) s=1 ;
          else if (s>100) s=100 ; 
          
         /* now 20*log10(s) is SNR */
         s=4-1.5*log10(s) ;  /* oversubstraction coeficient osub */
       }
      else s=4 ; /* this is the precomputed osub for SNR=0 dB and worse */    
     /* now s holds the oversubstraction factor */ 
     
     s=1-sqrt(s * pn/(*bp)) ; /* gain q */
     
     /* This is the rectification rule. The alternate formula for
        gain (s) reflects the original article but as far as I can 
        hear, there is no difference if you make s=0.0 */ 
        
     if (s<0 || *ap*s*s < SUBF*SUBF * pn)  s=sqrt(pn*SUBF*SUBF/ *ap) ;  
        /* this subband would be too weak, use noise instead */

     /* note we actually compute the complex cojugate of the
        output spectrum, divided by frame length to allow using
        FFT instead of IFFT */
        
     cp->r *= s/WLEN ; cp->i *= -s/WLEN ;            
    }  
}

/* ********************************************************** */

static void work()
{ 
  /* This routine takes one input frame and produces one
     output, hopefully noiseless output frame */

  float *fp,*ap ; complex *qp ;

  /* convert from real to complex and on the way apply the 
     Hanning window */
  for(ap=inbuf,qp=tbuf,fp=hann;ap<&inbuf[WLEN];ap++,fp++,qp++)
     { qp->r = *fp * (*ap) ; qp->i = 0 ; }
  
  /* compute the FFT of the input frame */
  fft(tbuf,spec) ;
  
  /* compute the amplitudes and the smoothed power spectrum */
  smooth() ;  

  /* run the noise estimator - reads px and generates pmin */ 
  noise() ;
             
  /* and finally we can apply the filtration 
     this gives us the new spectrum */
  filter() ; 
  
  /* note that filter inverts imaginary part of spec */
  /* so that we can compute IFFT using FFT simply as: */
  fft(spec,tbuf) ;
  
  /* note we discard imaginary part of tbuf so we do not have
     to bother with it any more and that filter() also makes
     the necesarry division */
 
  /* And that's it. The result stays in tbuf */
}

/* ********************************************************** */

int martin_do()
{ /* This is the main function of the module. It is called whenever
     master wants next subframe. It takes care about assembling
     frames from subframes on input and overlapping windows on output */ 

  float *f ; complex *cp ;
  int r ;
  
  if (empty) /* input buffer empty ? */
    { empty=0 ; ;
      /* fill the input buffer */
      if ((r=getfloat(inbuf,WLEN))!=0) return(r) ;          
    }
   else /* input buffer is full */
    { /* discard the oldest part and move us forward */
      memmove(inbuf,inbuf+WSTEP,(WLEN-WSTEP)*sizeof(float)) ;
      /* and load the next chunk */
      if ((r=getfloat(inbuf+WLEN-WSTEP,WSTEP))!=0) return(r) ;
    }     
  
  /* move the accumulating buffer too */
  memmove(accbuf,accbuf+WSTEP,(WLEN-WSTEP)*sizeof(float)) ;
  /* zero the newly released part */
  for(f=accbuf+WLEN-WSTEP;f<&accbuf[WLEN];f++) *f=0.0 ;  
 
  /* now do the actual work ! */
  work() ;

  /* and add the results */
  for(f=accbuf,cp=tbuf;f<&accbuf[WLEN];) (*(f++)) += (cp++)->r ;
  
  /* And that's it. Wait till we are called again */
  return(0) ;
}  

/* ******************************************************** */ 
 
int martin_init(readftype *f,int *sizep,float **outbuf,void *info)
 { 
   /* This function is called at the very beginning to establish
      the cooperation. Namely we find out the function to call
      when we need more data (f) and tell them our prefered
      chunk size (sizep) and where we place the output data
      (outbuf) */
   
   int r ;
   float *fp ; 

   /* store the getfp function pointer and set chunk size 
      and outbuf */
   getfp=f ; *sizep=WSTEP ; *outbuf=accbuf ;
   
   /* set the flags */
   empty=1 ; first=1 ; mi=0 ;
   
  /* set the accumulating output buffer to zero */
   for(fp=accbuf;fp<&accbuf[WLEN];fp++) *fp=0.0 ;       

  /* set the smooothed signal power to zero */
   for(fp=px;fp<&px[WLEN];fp++) *fp=0.0 ;       
  
   /* precompute Hanning coeficients */
   pi=acos(-1.0) ; /* Silly we have to do this.. */
   for(r=0;r<WLEN;r++) 
     hann[r]=((float)WSTEP/WLEN)*(1-cos(2*pi*r/(WLEN-1))) ;
   
   
   /* initialize FFT */
   if ((r=initfft(POWER))!=0) return(r) ;

   return(0) ;
 }           

/* ******************************************************** */ 
 
int martin_close()
{ /* And this is to say goodbye. If we had any memory to 
     deallocate we would do it here. */ 

   closefft() ;
   return(0) ;
}  
