/* This is the main source file for program snr, 
   calculating SNR for 16-bit, 8kHz, mono raw sound files, assuming
   clean signal and detector output are known. It prints the resulting
   mean total SNR and mean SNR for voiced parts. 
   
   For info, try typing snr with no arguments.
   
   Jan Kybic, February-April 1996
*/

#include <stdio.h>
#include <math.h>
 
#define WLEN 256      /* FFT window length */
#define POWER 8       /* 2^POWER=WLEN */
#define WSTEP 128     /* FFT window shift step size */
#define MINEN 1e-10   /* minimum energy per window */
#define MINRAT 1e-300 /* minimum ratio between signal and noise energy */
#define MAXRAT 1e+300

#define INT short int



/* calculates sum of squares of WSTEP INT's long vector */
static float vectene(INT *p) 
{ float e ;
  INT *q ;

  for(e=0,q=p;q<p+WSTEP;q++) e+=(*q) * (*q) ;
  return e ;
}


void main(int argc,char *argv[])
{ 
  float en1,en2,en ;  /* energy of voice+noise signal */
  float ec1,ec2,ec ;  /* energy of clean voice signal */
  double ratio ;       /* energy ratio */
  float snrt,snrv ;   /* total and voice SNR */
  int det ; 		/* current detector output */
  FILE *infn,*infc,*infd,*outfs ;  
  static INT ibufn[WSTEP],ibufc[WSTEP] ; /* integer input buffers */
  INT     *ipn,*ipc ; 

  int count,countv ; /* count of all and voiced windows */

  puts("snr version 1.0, SNR calculation") ;
  
  if (argc!=4 && argc!=5)
   { puts("usage: snr noisy_sig.bin clean_sig.bin voice.det [snr.dat]") ;
     puts(" noisy_sig ... measured signal") ;
     puts(" clean_sig ... corresponding clean signal") ;
     puts(" voice.det ... detector output") ; 
     puts(" snr.dat   ... SNR for each segment, binary form ") ;
     puts("Signals are binary, mono, 16bit, signed, host ordering.") ;
     puts("We assume 256 samples/window, 128 samples frame shift") ;
     return ;
   }
   
  if ((infn=fopen(argv[1],"r"))==NULL)
   { printf("snr: !!! Cannot open %s for reading.\n",argv[1]) ;
     return ; }

  if ((infc=fopen(argv[2],"r"))==NULL)
   { printf("snr: !!! Cannot open %s for reading.\n",argv[2]) ;
     return ; }

  if ((infd=fopen(argv[3],"r"))==NULL)
   { printf("snr: !!! Cannot open %s for reading.\n",argv[3]) ;
     return ; }
     
  if (argc==4) outfs=NULL ;
   else
  if ((outfs=fopen(argv[4],"w"))==NULL)
   { printf("snr: !!! Cannot open %s for writing.\n",argv[4]) ;
     return ; }

  en1=ec1=0.0 ; snrt=snrv=0.0 ; count=countv=0 ;
  /* read all windows */
  while (fread(ibufn,sizeof(INT),WSTEP,infn)==WSTEP)
    { 
      if (fread(ibufc,sizeof(INT),WSTEP,infc)!=WSTEP)
        { puts("snr: !!! Cannot read clean signal") ;
          return ;
        }
      if (count==0) det=0 ;/* The very first frame - no detector output */
        else
         if (fscanf(infd,"%u",&det)!=1)
          { puts("snr: !!! Warning - detector output too short") ;
            det=0 ;
          }
      
      if (det!=0 && det!=1) 
        { puts("snr: !!! Unexpected detector output (not 0 or 1)") ;
          return ;
        }
        
      /* get the noise signal by substracting clean sig from the mix */
      for(ipn=ibufn,ipc=ibufc;ipn<ibufn+WSTEP;ipn++,ipc++) (*ipn)-=(*ipc) ; 
      en2=vectene(ibufn) ; en=en1+en2 ; en1=en2 ;
      ec2=vectene(ibufc) ; ec=ec1+ec2 ; ec1=ec2 ;
              
      /* calculate the energy ratio and limit it */        
      if (en<MINEN) ratio=MAXRAT ; 
       else if (ec<MINEN) ratio=MINRAT ;
         else  ratio=(double)ec/en ;  
          
      /* convert to dB */    
      ratio=10*log10(ratio) ;
#if 0
      printf("%6.2g %6.2g %6.2f\n",en2,ec2,ratio) ;
#endif      

      if (outfs) fwrite(&ratio,sizeof(double),1,outfs) ;
          
      count++ ; snrt+=ratio ;
      if (det) { countv++ ; snrv+=ratio ; }
		      
    }
  if (count>0) printf("global SNR=%.1f  ",snrt/count) ;
     else { puts("Signal too short, SNR not available") ;
            return ; }

  if (countv>0) printf("voiced segments SNR=%.1f\n",snrv/countv) ;
     else { puts(" no voiced segments") ;
          }
             		      
  fprintf(stderr,"%.1f : %.1f\n",snrt/count,snrv/countv) ;		      
  fclose(infn) ; fclose(infc) ; fclose(infd) ; fclose(outfs) ;    
}     