/* This is a source file for 'C' implemetations */
/* Noisy speech enhancement algorithm. 
   Two step spectral subtraction, full-wave rectification */
/* Jan Kybic */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fft.h"
#include "so1ff.h"

#define WLEN 256     /* FFT window length */
#define POWER 8      /* 2^POWER=WLEN */
#define WSTEP 128     /* FFT window shift step size */

#define PAUSE 1       /* initial unvoiced segment length in seconds */
#define HZ    8000    /* sampling frequency */

#define P  .8	     /* smoothing factor */
#define MINAMP 1e-20 /* minimal amplitude to keep phase */

#define THRESHOLD 1.5 /* threshold for the energy detector */
#define K 0.2         /* adaptation constant */

#define FULLWAVE 0    /* just one way rectification */


static readftype *getfp ;    /* pointer to the reading function */
static float inbuf[WLEN] ;   /* input buffer */
static complex spec[WLEN] ;  /* input/output spectrum */
static float ax[WLEN] ;      /* current spectral energy spectrum */
static float dx[WLEN] ;      /* noise energy spectrum estimate */
static float ay[WLEN] ;      /* signal after first subtraction */
static float rx[WLEN] ;      /* residual noise energy spectrum */
static float sx[WLEN] ;      /* output signal energy spectrum */
static float accbuf[WLEN] ; /* accumulating buffer */
static complex tbuf[WLEN] ; /* here the frame output goes */
static float hann[WLEN] ; /* precomputed Hanning window coeficients */
static float pi ;         /* 3.1415... */
static float energy ;     /* energy level for the detector */ 
static char empty ;       /* flags */
static FILE *detf ;       /* file containing detector output */
static struct so1ff_info *info ;
static unsigned long framen ;

/* ********************************************************** */

/* reads num samples to dest using getfp 
   num must be multiple of chunk size WSTEP */ 
   
static int getfloat(float *dest,int num) 
{ int r ;  
  while(num>0)
   { /* read from the input */
     if ((r=(*getfp)(dest))!=0) return(r) ;
     /* and in(de)crement the counters */
     dest+=WSTEP ; num-=WSTEP ;
   }
  return(0) ;    
}   

/* ********************************************************** */

static void currene()
{ /* spec -> ax, px */ 
  /* compute the current spectral energy */
  
  float *ap ; complex *cp ;
  
  for(cp=spec,ap=ax;cp<&spec[WLEN];cp++,ap++)
    { *ap=(cp->r)*(cp->r) + (cp->i)*(cp->i) ;
    }
}

/* ********************************************************** */

/* energy speech detector, same as in enedet.c */

static int energdet()
{ float e ; float *ap ;

  e=0 ; /* calculate total energy */
  for(ap=ax;ap<ax+WLEN;ap++) e+=*ap ;
  
  if (energy<0) { energy=e ; return 0 ; }
  
  if (e<=THRESHOLD*energy) { energy=(1-K)*energy+K*e ;
                             return 0 ;
                           }
  return 1 ;
}                                    


static int speechd() /* Call the appropriate detector */
{ int r ;

   if (detf==NULL) 
     if (info->detfile) { if (framen++<(PAUSE*HZ)/WSTEP) r=0 ;
                                               else  r=1 ;
                        }                       
                else    r=energdet() ;
    else {  r=0 ; fscanf(detf,"%d",&r) ; }
  return r ; 
}   

/* ********************************************************** */


static void subst(float *from,float *to,float *noise,int det)
{ 
  /* Here we implement the spectral subtraction */
  
  float *ap,*bp ; 
  float s ;
  
  /* If there is not a voice, update noise spectrum estimate */
  if (det==0) 
    for(ap=from,bp=noise;ap<from+WLEN;ap++,bp++)
      *bp=(1-P)*(*bp)+P*(*ap) ;
     
     
  /* go through the whole spectrum and subtract */
  for(ap=from,bp=to;ap<from+WLEN;ap++,bp++,noise++)
   { s=*ap - *noise ; 
     if (info->fullwave) *bp=s>0 ? s : -s ;
          else           *bp=s>0 ? s : 0 ;
     
   }  
}

/* ********************************************************** */


static void phase()
{ 
  /* Here we add the phase information and 
    calculate the complex conjugate of the output spectrum divided
    by the frame length. */
  
  float *ap,*sp ; complex *cp ;
  float s ;
  
  /* go through the whole spectrum and multiply */
  for(ap=ax,sp=sx,cp=spec;ap<ax+WLEN;ap++,sp++,cp++)
   { if (*ap<MINAMP) { cp->r=*sp / WLEN ; cp->i=0 ; 
                     }
                else { s=sqrt((*sp)/(*ap)) ;
                       cp->r *= s/WLEN ; cp->i *= -s/WLEN ;
                     }
   }                     
}

/* ********************************************************** */

static void work()
{ 
  /* This routine takes one input frame and produces one
     output, hopefully noiseless output frame */

  float *fp,*ap ; complex *qp ; int det ;

  /* convert from real to complex and on the way apply the 
     Hanning window */
  for(ap=inbuf,qp=tbuf,fp=hann;ap<&inbuf[WLEN];ap++,fp++,qp++)
     { qp->r = *fp * (*ap) ; qp->i = 0 ; }
  
  /* compute the FFT of the input frame */
  fft(tbuf,spec) ;
  
  /* compute the amplitude spectrum */
  currene() ;  

  /* run the speech detector */ 
  det=speechd() ;

  if (info->twice)             
    { /* we can apply the spectral subtraction now, 
       this gives us the new spectrum: ax -> ay */
      subst(ax,ay,dx,det) ; 
  
      /* apply the filtration once more: ay -> sx */
     subst(ay,sx,rx,det) ;
    }
   else subst(ax,sx,rx,det) ;   
     /* add the phase information: sx -> spec */
  phase() ;
  
  /* note that phase inverts imaginary part of spec */
  /* so that we can compute IFFT using FFT  */

  fft(spec,tbuf) ;
  
  /* note we discard imaginary part of tbuf so we do not have
     to bother with it any more and that filter() also makes
     the necesarry division */
 
  /* And that's it. The result stays in tbuf */
}

/* ********************************************************** */

int so1ff_do()
{ /* This is the main function of the module. It is called whenever
     master wants next subframe. It takes care about assembling
     frames from subframes on input and overlapping windows on output */ 

  float *f ; complex *cp ;
  int r ;
  
  if (empty) /* input buffer empty ? */
    { empty=0 ; ;
      /* fill the input buffer */
      if ((r=getfloat(inbuf,WLEN))!=0) return(r) ;          
    }
   else /* input buffer is full */
    { /* discard the oldest part and move us forward */
      memmove(inbuf,inbuf+WSTEP,(WLEN-WSTEP)*sizeof(float)) ;
      /* and load the next chunk */
      if ((r=getfloat(inbuf+WLEN-WSTEP,WSTEP))!=0) return(r) ;
    }     
  
  /* move the accumulating buffer too */
  memmove(accbuf,accbuf+WSTEP,(WLEN-WSTEP)*sizeof(float)) ;
  /* zero the newly released part */
  for(f=accbuf+WLEN-WSTEP;f<&accbuf[WLEN];f++) *f=0.0 ;  
 
  /* now do the actual work ! */
  work() ;

  /* and add the results */
  for(f=accbuf,cp=tbuf;f<&accbuf[WLEN];) (*(f++)) += (cp++)->r ;
  
  /* And that's it. Wait till we are called again */
  return(0) ;
}  

/* ******************************************************** */ 
 
int so1ff_init(readftype *f,int *sizep,float **outbuf,void *infoadr)
 { 
   /* This function is called at the very beginning to establish
      the cooperation. Namely we find out the function to call
      when we need more data (f) and tell them our prefered
      chunk size (sizep) and where we place the output data
      (outbuf) */
   
   int r ;
   float *fp ; 
   
   info=infoadr ;

   /* store the getfp function pointer and set chunk size 
      and outbuf */
   getfp=f ; *sizep=WSTEP ; *outbuf=accbuf ;
   
   /* set the flags */
   empty=1 ; energy=-1.0 ;
   
   framen=0 ;
   
   /* set the accumulating output buffer and noise estimates to zero */
   for(fp=accbuf;fp<&accbuf[WLEN];fp++) 
     { *fp=*((fp-accbuf)+dx)=*((fp-accbuf)+rx)=0.0 ; }        

   /* precompute Hanning coeficients */
   pi=acos(-1.0) ; /* Silly we have to do this.. */
   for(r=0;r<WLEN;r++) 
     hann[r]=((float)WSTEP/WLEN)*(1-cos(2*pi*r/(WLEN-1))) ;

   detf=NULL ;   
   if (info->detfile && *info->detfile)
                if ((detf=fopen(info->detfile,"r"))==NULL)
                  { perror("so1ff: Cannot open detector file") ;
                    return(-1) ;
                  }  
                
   /* initialize FFT */
   if ((r=initfft(POWER))!=0) return(r) ;

   return(0) ;
 }           

/* ******************************************************** */ 
 
int so1ff_close()
{ /* And this is to say goodbye. If we had any memory to 
     deallocate we would do it here. */ 

   if (info->detfile && *info->detfile) fclose(detf) ;
   closefft() ;
   return(0) ;
}  
