/* C implementation of

% Toeplitz Structured Total Least Squares Algorithm
% 'x' containts measurements corrupted by white noise
% on output 'a' contains the coefficients of an AR system of the order 'n'
% generating output 'y', while (x-y) is minimised in Euclidean sense
% y(i)=a(1)*y(i-1)+a(2)*y(i-2)+...+a(n)*y(i-n)
% a=arg min sum((y(i)-x(i))^2)
%
% Uses weighting to get unconstrained problem that solves using (conjugate) 
% gradient method

Jan Kybic
*/

#ifndef MAKING_MEX
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "kalmse.h"

#define SAFEMALLOC(ptr,a,b) ptr=safemalloc((a)*(b)*sizeof(double)) ;
#define SAFEMALLOCINT(ptr,a,b) ptr=safemalloc((a)*(b)*sizeof(int)) ;
#define mxFree(a) safefree(a)

#define mexPrintf printf
#endif

#ifdef MAKING_MEX
#include "burg.c" /* the Burg algorithm */
#endif

#define OMEGA (1e2)    /* the weight constant */

#define MINITER (5)    /* minimum number of iterations */
#define MAXINITER (50) /* maximum number of iterations for the internal loop */
#define SML (1e-4)     /* 'small number', minimum relative step size */
#define GLD (0.38197)  /* golden ratio for minimum finding */
#define PREC (1e-10)    /* the overall precision */
#define TOL  (1e-6)    /* relative error for minimum finding */
#define ABSTOL (1e-20) /* absolute error for minimum finding */
#define EXPMAG (10)    /* maximum expected magnitude of the steps */
#define EPS    (1e-30) /* about machine precision */

#ifndef MAX
#define MAX(a,b) ((a)>(b) ? (a) : (b))
#define MIN(a,b) ((a)>(b) ? (b) : (a))
#endif

/* global variables for faster access */
static double *g_x ;
static int g_n,g_T ;
static double *g_z,*g_dz,*g_ztemp,*g_ee ;

/* the  z vector consists of [y a] */

double crit(double *z)
     /* evaluates the criterium at point z, sets g_ee */ 
{ int i,j ;
  double sum,sum1,tmp,*a ;
  
  a=z+g_T ;
  for(sum=0.0,i=0;i<g_T;i++)
    { tmp=z[i]-g_x[i] ; sum+=tmp*tmp ; }

  for(sum1=0.0,i=g_n;i<g_T;i++)
    { g_ee[i]=z[i] ;
      for(j=0;j<g_n;j++) g_ee[i]-=a[j]*z[i-j-1] ;
      sum1+=g_ee[i]*g_ee[i] ;
    }
  sum+=sum1*OMEGA ;
  return sum ;
}

void grad(double *z)
     /* calculates the gradient and stores it in g_dz */
{ int i,j,k ;
  double sum,*a ;
  a=z+g_T ;  

  for(i=0;i<g_T;i++)
    { k=MIN(g_n,g_T-i-1) ;
      for(sum=g_ee[i],j=0;j<k;j++) sum-=g_ee[i+j+1]*a[j] ;
      g_dz[i]=2*((z[i]-g_x[i])+OMEGA*sum) ;
    }
  for(i=0;i<g_n;i++)
    { sum=0.0 ;
      for(j=g_n;j<g_T;j++) sum-=g_ee[j]*z[j-i-1] ;
      g_dz[i+g_T]=2*OMEGA*sum ;
    }
}

double eval_along(double *dz,double alpha)
     /* evaluates the criteria at -alpha*dz+g_z */
{ int i ; 
  for(i=0;i<g_n+g_T;i++) g_ztemp[i]=g_z[i]-alpha*dz[i] ;
  return crit(g_ztemp) ;
}


double minfind(double *dz,double J0)
     /* finds and stores to g_z the location of the minimum of the
	criteria on the line from the old value of g_z along the line given by
	dz needs to know J0(g_z). Returns the new minimum J */
{
  double alpha0,alpha1,alpha,J1,J,sum,newJ,newalpha ;
  double dif1,dif2,dif3,dif4,den ;
  int i ;

  /* scale dz appropriately - a bit of a hack, really */
  sum=0.0 ;
  for(i=0;i<g_n+g_T;i++) sum+=dz[i]*dz[i] ;
  sum=sqrt(sum)/EXPMAG ;
  if (sum>1) 
    { sum=1/sum ;
      for(i=0;i<g_n+g_T;i++) dz[i]*=sum ;
    }

  /* try full step */
  J1=eval_along(dz,alpha1=1.0) ;
  alpha0=0.0 ;
  
  if (J1<J0) { /* function decreases -> find when it starts to increase */
    J=J0 ;
    while (J1<J) {
      J=J1 ; alpha=alpha1 ;
      alpha1*=10 ;
      J1=eval_along(dz,alpha1) ;
    } }
  else {
    /* function increases -> go backward until it decreases */
    alpha=0.5 ;
    J=eval_along(dz,alpha) ;
    while (J>J0) {
      J1=J ; alpha1=alpha ;
      alpha*=0.1 ;
      J=eval_along(dz,alpha) ;
    }
  }

  /* now the minimum is bracketted between alpha0<alpha<alpha1 with
  corresponding J0,J,J1 where J<J0, J<J1 */

  for(i=0;i<MAXINITER;i++) { 
    if (i>0 && fabs(alpha1-alpha0)<MAX(TOL*alpha,ABSTOL)) 
      break ; /* success */
    dif1=alpha-alpha1 ; dif2=J-J0 ; dif3=alpha-alpha0 ; dif4=J-J1 ;
    den=dif1*dif2-dif3*dif4 ;
    if (fabs(den)<EPS) /* colinear points */
      if (J1>J0)
	newalpha=alpha*(1-GLD)+GLD*alpha0 ;
      else
	newalpha=alpha1*GLD+(1-GLD)*alpha ;
    else { 
      newalpha=alpha-0.5*(dif1*dif1*dif2-dif3*dif3*dif4)/den ;
      newalpha=MAX(MIN(newalpha,alpha1-SML*(alpha1-alpha0)),
		   alpha0+SML*(alpha1-alpha0)) ;
    }
    /* we have the new alpha now */
    newJ=eval_along(dz,newalpha) ;
    if (newJ>J)
      if (newalpha<alpha) {
	alpha0=newalpha ; J0=newJ ;
      }
      else {
	alpha1=newalpha ; J1=newJ ;
      }
    else
      if (newalpha<alpha) {
	alpha1=alpha ; J1=J ; alpha=newalpha ; J=newJ ; 
      }
      else {
	alpha0=alpha ; J0=J ; alpha=newalpha ; J=newJ ;
      }
  } /* for i */
  
  J=eval_along(dz,alpha) ;
  memmove(g_z,g_ztemp,(g_n+g_T)*sizeof(double)) ;
  
  return J ;
}

void tstlsw(double *xp,double *ap,int n,int T,int maxiter)
{ int i,j ;
  double *g,*h,J,J0,gamma1,gamma ;

  SAFEMALLOC(g_z,1,n+T)
  SAFEMALLOC(g_dz,1,n+T)
  SAFEMALLOC(g_ztemp,1,n+T)
  SAFEMALLOC(g_ee,1,n+T)
  SAFEMALLOC(g,1,n+T)
  SAFEMALLOC(h,1,n+T)


  /* set the global variables */
  g_x=xp ; g_n=n ; g_T=T ;

  burg(xp,g_z+T,n,T) ;
  memmove(g_z,xp,T*sizeof(double)) ;

  for(i=0;i<T+n;i++) g_ee[i]=0.0 ;

  J=crit(g_z) ;

#ifdef DEBUG
  mexPrintf("maxiter=%d\n",maxiter) ;
#endif

  for (i=0;i<maxiter;i++) {

#ifdef DEBUG
    mexPrintf("i=%d J=%g\n",i,J) ;
#endif
    J0=J ; grad(g_z) ;
    if (i==0) {
      memmove(g,g_dz,(T+n)*sizeof(double)) ;
      memmove(h,g_dz,(T+n)*sizeof(double)) ;
    }
    else {
      for(gamma1=0.0,j=0;j<n+T;j++) gamma1+=g[j]*g[j] ;
      for(gamma=0.0,j=0;j<n+T;j++) gamma+=(g_dz[j]-g[j])*g_dz[j] ;
      gamma/=gamma1 ;
      memmove(g,g_dz,(T+n)*sizeof(double)) ;
      for(j=0;j<n+T;j++) h[j]=h[j]*gamma+g[j] ; 
    }

    memmove(g_dz,h,(n+T)*sizeof(double)) ; 
    /* because g_dz will get changed in minfind */
     
    J=minfind(g_dz,J0) ;
    if (i>MINITER & (J0-J)/J<PREC) break ;
  }

  memmove(ap,g_z+T,n*sizeof(double)) ;  

  mxFree(g_z) ; mxFree(g_dz) ; mxFree(g_ztemp) ; mxFree(g_ee) ;
  mxFree(g) ; mxFree(h) ;
}
