function [a,y]=tstls(x,n) 
% Toeplitz Structured Total Least Squares Algorithm
% 'x' containts measurements corrupted by white noise
% on output 'a' contains the coefficients of an AR system of the order 'n'
% generating output 'y', while (x-y) is minimised in Euclidean sense
% y(i)=a(1)*y(i-1)+a(2)*y(i-2)+...+a(n)*y(i-n)
% a=arg min sum((y(i)-x(i))^2)
%
% Based on STLNB algorithm from S. Van Huffel, Haesun Park, J. Ben Rosen: 
% 'Formulation and Solution of Structured Total Least Norm Problems for
% Parameter Estimation', IEEE Transactions on Signal Processing, 1996
%
% Jan Kybic, 1997
%
% Usage: [a y]=tstls(x,n)


x=x(:) ;      % make x a column vector
T=length(x) ; % number of available samples
m=T-n ;       % number of equations

prec=1e-6 ;   % desired precision
maxiter=50 ; % maximum number of iterations

% construct the left hand side matrix
A=zeros(m,n) ;
for i=1:n, A(:,i)=x(i:i+m-1) ; end ;

b=x(n+1:T) ; % right hand side

% initialize the iterative procedure
omega=1e7 ;
E=zeros(m,n) ;
alpha=zeros(T,1) ; 
a=A\b ;

Xe=zeros(m,T-1) ;
left=zeros(n+1+T) ; right=zeros(1,n+T) ;
Fhat=[zeros(m,n) [zeros(1,m-1) ; eye(m-1)]] ;

lsrh=0 ;

for i=1:maxiter,
  i
  % Construct the E matrix
  for i=1:n, E(:,i)=alpha(i:i+m-1) ; end ;

  % Construct the Xe matrix 
  for j=1:m, Xe(j,j:j+n-1)=a' ; end ;
  
  rhat=b+alpha(n+1:T)-(A+E)*a ; srh=sum(rhat.^2)
  if abs(lsrh-srh)<prec, 
    i=maxiter ;
  end ;
  lsrh=srh ;
  
  % Solve the LS minimalization problem
  left=[ omega*(Xe-Fhat) [-omega ; zeros(m-1,1)] omega*(A+E) ; ...
    eye(T) zeros(T,n) ] ;
  right=[omega*rhat ; -alpha ] ;
  delta=left\right ;
  
  % Update variables
  a=a+delta(T+1:T+n) ;
  alpha=alpha+delta(1:T) ;
end ;

y=x+alpha ;
a=a(n:-1:1)' ; % get the usual order