function [a,y]=tstls1(x,n) 
% Toeplitz Structured Total Least Squares Algorithm
% 'x' containts measurements corrupted by white noise
% on output 'a' contains the coefficients of an AR system of the order 'n'
% generating output 'y', while (x-y) is minimised in Euclidean sense
% y(i)=a(1)*y(i-1)+a(2)*y(i-2)+...+a(n)*y(i-n)
% a=arg min sum((y(i)-x(i))^2)
%
% Use quasi Newton iterative method to solve Lagrange equations
% It should have quadratic convergence.
% The implementation tries to be quick but ignores some numerical issues on
% the premises of a restricted magnitude range of the quantities involved
%
% Jan Kybic, 1997
%
% Usage: [a y]=tstls1(x,n)

x=x(:) ;      % make x a column vector
T=length(x) ; % number of available samples
m=T-n ;       % number of equations

prec=1e-6 ;   % desired precision
maxiter=50 ;  % maximum number of iterations
maxiniter=20 ;% maximum inner number of iterations

% initial guess
a=burg1(x,n)' ;

% scale the problem
scale=sum(abs(x))/T ;
% scale=1 ;
x=x/scale ;

% initial value of z= [ a ; y ; lambda(n+1:T)]
z=[a ; x ; ones(T-n,1) ];

fval=0 ;

for i=1:maxiter,
  f=eval_right_side(z,n,T,x) ;
  oldfval=fval ; fval=sum(f.*f)
  if i>1 & fval>oldfval,
    z=z+dz ;      % undo the last step
    for j=1:maxiniter,
      dz=0.1*dz ; % make the step smaller 
      f=eval_right_side(z-dz,n,T,x) ;
      fval=sum(f.*f) ;     
      if fval<oldfval, 
	disp([num2str(j) ' smaller steps taken']) ;
	break ;
      end ;
    end ;
    if fval>=oldfval,
      disp('Divergence detected') ;
      return ;
    end ;
    z=z-dz ; % take the step
  end ;
  
  if fval<prec, % desired precision reached ? 
    a=z(1:n) ;
    y=z(n+1:T+n)*scale ;
    return ;
  end ;
  
% dz=slow_newton(z,f,n,T) ; % this should give the same results
  dz=newton_step(z,f,n,T) ; 
  z=z-dz ;
end ;

% all the maxiter iterations taken
a=z(1:n) ;
y=z(n+1:T+n)*scale ;

% --------------------------------------------------------------------
function f=eval_right_side(z,n,T,x)
% Evaluates the Lagrangian f=F(z) at point z
% z=[y(1)..y(T) a(1)..a(n) lambda(n+1)..lamda(T)] is a column vector
% f has the dimension 2*T x 1
f=zeros(2*T,1) ;
y=z(n+1:n+T) ;
a=z(1:n) ;
lambda=[zeros(n,1) ; z(T+n+1:2*T) ; zeros(n,1)] ;
% equations I
for t=1:T,
  f(t)=2*(y(t)-x(t)) + [1 -a(1:n)']*lambda(t:t+n) ;
end ;
% equations II
for k=1:n,
  f(T+k)= - lambda(n+1:T)' * y(n+1-k:T-k) ;
end ;  
% equations III
for t=n+1:T,
  f(T+t)=y(t)-a(1:n)'*y(t-1:-1:t-n) ;
end ;  

% --------------------------------------------------------------------
function dz=newton_step(z,f,n,T) 
% given the point z and the corresponding Lagrangian f=F(z),
% find the Newton step dz=inv(dF/dz)*f taking advantage of the 
% Jacobian (dF/dz) matrix sparsity and structure
y=z(n+1:n+T) ;
a=z(1:n) ;
lambda=[zeros(n,1) ; z(T+n+1:2*T)] ;

% generate P,M,g from III such that 
% P*[dy(1)..dy(n)]'+M*[da(1)..da(n)]'+g=[dy(1)..dy(T)]'
P=[eye(n) ; zeros(T-n,n) ] ;
M=zeros(T,n) ;
g=zeros(T,1) ;
for t=n+1:T,
  P(t,:)=a(n:-1:1)'*P(t-n:t-1,:) ;
  g(t)=f(t+T)+a(n:-1:1)'*g(t-n:t-1) ;
  M(t,:)=y(t-1:-1:t-n)'+a(n:-1:1)'*M(t-n:t-1,:) ;
end ;  

% generate (from Ib) Q,R,r such that 
% Q*[dy(1)..dy(n)]'+R*[da(1)..da(n)]'+r=[dlambda(n+1)..dlambda(T)]'
Q=zeros(T-n,n) ;
r=zeros(T-n,1) ;
Q(T-n,:)=-2*P(T,:) ;
r(T-n)=f(T)-2*g(T) ;
R=zeros(T-n,n) ;
R(T-n,:)=-2*M(T,:) ;
for t=T-1:-1:n+1 ;
  k=min(T-t,n) ;
  Q(t-n,:)=-2*P(t,:)+a(1:k)'*Q(t-n+1:t-n+k,:) ;
  r(t-n)=f(t)-2*g(t)+a(1:k)'*r(t-n+1:t-n+k) ;
  mx=min(t+n,T) ;
  R(t-n,1:mx-t)=lambda(t+1:mx)' ;
  R(t-n,:)=R(t-n,:)-2*M(t,:)+a(1:k)'*R(t-n+1:t-n+k,:) ;
end ;  

% convert equations II to a form A*[a(1)..a(n) y(1)..y(n)]'=c
A=zeros(n,2*n) ;
c=zeros(n,1) ;
for i=1:n,
  c(i)=f(i+T)+lambda(n+1:T)'*g(n-i+1:T-i) ;
  c(i)=c(i)+y(n+1-i:T-i)'*r(1:T-n) ;
  A(i,1:n)=-y(n+1-i:T-i)'*R-lambda(n+1:T)'*M(n-i+1:T-i,:) ;
  A(i,n+1:2*n)=-y(n+1-i:T-i)'*Q-lambda(n+1:T)'*P(n-i+1:T-i,:) ;
end ;  

% convert equations Ia to B*[a(1)..a(n) y(1)..y(n)]'=d
B=zeros(n,2*n) ;
d=zeros(n,1) ;
for i=1:n,
  B(i,1:n)=-lambda(i+1:i+n)'-a(n-i+1:n)'*R(1:i,:) ;
  B(i,n+1:2*n)=2*P(i,:)-a(n-i+1:n)'*Q(1:i,:) ;
  d(i)=f(i)-2*g(i)+a(n-i+1:n)'*r(1:i) ;
end ;

% solve
dz=zeros(2*T,1) ;
h=[A ; B] \ [c ; d] ;
dz(1:2*n)=h ; % da dy(1:n)
a=h(1:n) ; y=h(n+1:2*n) ;

% backsubstitute to get the rest of the variables
dz(n+1+n:T+n)=P(n+1:T,:)*y+M(n+1:T,:)*a+g(n+1:T) ;       % dy(n+1:T)
dz(T+n+1:2*T)=Q*y+R*a+r ; % lambda(n+1:T)

% ------------------------------------------------------------------
function dz=slow_newton(z,f,n,T)
% Performs slow Newton step computation, for test purposes
% It should give exactly the same result as newton_step

J=zeros(2*T,2*T) ;
y=z(n+1:n+T) ;
a=z(1:n) ;
lambda=[zeros(n,1) ; z(T+n+1:2*T) ; zeros(n,1)] ;

for i=1:T,
  J(i,1:n)=-lambda(i+1:i+n)' ;
  J(i,i+n)=2 ;
end ;  
for i=n+1:T,
  J(i-n:i,i+T)=[-a(n:-1:1) ; 1] ;
end ;  
for i=1:n,
  J(i+T,2*n+1-i:n+T-i)=-lambda(n+1:T)' ;
  J(i+T,n+T+1:2*T)=-y(n-i+1:T-i)' ;
end ;  
for i=n+1:T,
  J(i+T,i:i+n)=[-a(n:-1:1)' 1] ;
  J(i+T,1:n)=-y(i-1:-1:i-n)' ;
end ;  

dz=J\f ;
