function [a,y]=tstls3(x,n) 
% Toeplitz Structured Total Least Squares Algorithm
% 'x' containts measurements corrupted by white noise
% on output 'a' contains the coefficients of an AR system of the order 'n'
% generating output 'y', while (x-y) is minimised in Euclidean sense
% y(i)=a(1)*y(i-1)+a(2)*y(i-2)+...+a(n)*y(i-n)
% a=arg min sum((y(i)-x(i))^2)
%
% Use quasi Newton iterative method to solve Lagrange equations
% It should have quadratic convergence.
% The implementation tries to be quick but ignores some numerical issues on
% the premises of a restricted magnitude range of the quantities involved
%
% This version tries to find minimum on backtracking
% It also tries to find a good initial estimate
%
% Jan Kybic, 1997
%
% Usage: [a y]=tstls3(x,n)

x=x(:) ;      % make x a column vector
T=length(x) ; % number of available samples
m=T-n ;       % number of equations

prec=1e-6 ;   % desired precision
maxiter=15000 ;  % maximum number of iterations
maxiniter=10 ;% maximum inner number of iterations

% initial guess of a
a=burg1(x,n)' ;

% scale the problem
scale=sum(abs(x))/T ;
% scale=1 ;
x=x/scale ;

y=estimate_y(a,x,n,T) ;
% [a y]=tstlsw(x,n) ;


z=[a ; y ; zeros(T-n,1)];

% z=estimate_lambda2(z,n,T,x) ;

% initial value of z= [ a ; y ; lambda(n+1:T)]
% z=[a ; y ; lambda];
% z=[a ; y ; 0*ones(T-n,1)];

fval=0 ;
% constants for the backtracking
alphamin=0.01 ;
alphamax=0.5 ;
% alphafirst=0.38197 ; % golden ratio
alphafirst=0.01 ;

i=0 ; % number of evaluations
while i<maxiter,
  
  f=eval_right_side(z,n,T,x) ;
  i=i+1 ;
  oldfval=fval ; fval=sum(f.*f) ;
  
  if i>1 & fval>oldfval,        % if not approaching the minimum
    z=z+dz ;                    % undo the last step
    fval1=fval ; alpha1=1 ;     % save the full step result
    % try smaller step
    f=eval_right_side(z-alphafirst*dz,n,T,x) ;
    i=i+1 ;
    fval=sum(f.*f) ; alpha2=alphafirst ;
    disp(['Evaluation ' num2str(i) ' fval=' num2str(fval)]) ;
      
    % the last two attempts are always in fval, fval1 with corresponding
    % alpha2,alpha1. alpha is the new value
    for j=1:maxiniter,
      if fval<oldfval, % success
	break ;
      end ;
      % find the new alpha
      if fval>fval1,
	% the function is behaving strangely - be conservative
	disp('Strange behaviour') ;
	alpha=alphamin*alpha2 ;
      else
	% find new alpha by parabolic approximation
	dif1=fval-fval1 ; dif2=fval-oldfval ; dif3=alpha2-alpha1 ;
	den=alpha2*dif1-dif3*dif2 ;
	if abs(den)<eps, % colinear points
	  alpha=alphamin*alpha2 ;
	else 
	  alpha=alpha2-0.5*(alpha2^2*dif1-dif3^2*dif2)/den ;
	  % force into bounds
	  alpha=max([min([alpha alphamax*alpha2]) alphamin*alpha2]) ; 
	end ;  
      end ; % of if fval>fval1
      % here we have the new alpha
      alpha1=alpha2 ; fval1=fval ; alpha2=alpha ;
      f=eval_right_side(z-alpha*dz,n,T,x) ;
      i=i+1 ;
      fval=sum(f.*f) ;
      disp(['Evaluation(backtracking)' num2str(i) ' fval=' num2str(fval) ]) ;
    end ; % of for j

    if fval>=oldfval,
      disp('Divergence detected') ;
      % just return the best values so far 
      break ;
    end ;
    z=z-dz*alpha2 ; % take the step, finally 
  end ; % of if fval>oldfval, (not approaching miminum)
    

  disp(['Evaluation(main) ' num2str(i) ' fval=' num2str(fval)]) ;
  if fval<prec, % desired precision reached ? 
    break ;
  end ;
  
% dz=slow_newton(z,f,n,T) ; % this should give the same results
  dz=newton_step(z,f,n,T) ; 
  z=z-dz ;
end ;

% all the maxiter iterations taken or convergence attained
a=z(1:n) ;
y=z(n+1:T+n)*scale ;

% --------------------------------------------------------------------
function f=eval_right_side(z,n,T,x)
% Evaluates the Lagrangian f=F(z) at point z
% z=[y(1)..y(T) a(1)..a(n) lambda(n+1)..lamda(T)] is a column vector
% f has the dimension 2*T x 1
f=zeros(2*T,1) ;
y=z(n+1:n+T) ;
a=z(1:n) ;
lambda=[zeros(n,1) ; z(T+n+1:2*T) ; zeros(n,1)] ;
% equations I
for t=1:T,
  f(t)=2*(y(t)-x(t)) + [1 -a(1:n)']*lambda(t:t+n) ;
end ;
% equations II
for k=1:n,
  f(T+k)= - lambda(n+1:T)' * y(n+1-k:T-k) ;
end ;  
% equations III
for t=n+1:T,
  f(T+t)=y(t)-a(1:n)'*y(t-1:-1:t-n) ;
end ;  

% --------------------------------------------------------------------
function dz=newton_step(z,f,n,T) 
% given the point z and the corresponding Lagrangian f=F(z),
% find the Newton step dz=inv(dF/dz)*f taking advantage of the 
% Jacobian (dF/dz) matrix sparsity and structure
y=z(n+1:n+T) ;
a=z(1:n) ;
lambda=[zeros(n,1) ; z(T+n+1:2*T)] ;

% generate P,M,g from III such that 
% P*[dy(1)..dy(n)]'+M*[da(1)..da(n)]'+g=[dy(1)..dy(T)]'
P=[eye(n) ; zeros(T-n,n) ] ;
M=zeros(T,n) ;
g=zeros(T,1) ;
for t=n+1:T,
  P(t,:)=a(n:-1:1)'*P(t-n:t-1,:) ;
  g(t)=f(t+T)+a(n:-1:1)'*g(t-n:t-1) ;
  M(t,:)=y(t-1:-1:t-n)'+a(n:-1:1)'*M(t-n:t-1,:) ;
end ;  

% generate (from Ib) Q,R,r such that 
% Q*[dy(1)..dy(n)]'+R*[da(1)..da(n)]'+r=[dlambda(n+1)..dlambda(T)]'
Q=zeros(T-n,n) ;
r=zeros(T-n,1) ;
Q(T-n,:)=-2*P(T,:) ;
r(T-n)=f(T)-2*g(T) ;
R=zeros(T-n,n) ;
R(T-n,:)=-2*M(T,:) ;
for t=T-1:-1:n+1 ;
  k=min(T-t,n) ;
  Q(t-n,:)=-2*P(t,:)+a(1:k)'*Q(t-n+1:t-n+k,:) ;
  r(t-n)=f(t)-2*g(t)+a(1:k)'*r(t-n+1:t-n+k) ;
  mx=min(t+n,T) ;
  R(t-n,1:mx-t)=lambda(t+1:mx)' ;
  R(t-n,:)=R(t-n,:)-2*M(t,:)+a(1:k)'*R(t-n+1:t-n+k,:) ;
end ;  

% convert equations II to a form A*[a(1)..a(n) y(1)..y(n)]'=c
A=zeros(n,2*n) ;
c=zeros(n,1) ;
for i=1:n,
  c(i)=f(i+T)+lambda(n+1:T)'*g(n-i+1:T-i) ;
  c(i)=c(i)+y(n+1-i:T-i)'*r(1:T-n) ;
  A(i,1:n)=-y(n+1-i:T-i)'*R-lambda(n+1:T)'*M(n-i+1:T-i,:) ;
  A(i,n+1:2*n)=-y(n+1-i:T-i)'*Q-lambda(n+1:T)'*P(n-i+1:T-i,:) ;
end ;  

% convert equations Ia to B*[a(1)..a(n) y(1)..y(n)]'=d
B=zeros(n,2*n) ;
d=zeros(n,1) ;
for i=1:n,
  B(i,1:n)=-lambda(i+1:i+n)'-a(n-i+1:n)'*R(1:i,:) ;
  B(i,n+1:2*n)=2*P(i,:)-a(n-i+1:n)'*Q(1:i,:) ;
  d(i)=f(i)-2*g(i)+a(n-i+1:n)'*r(1:i) ;
end ;

% solve
dz=zeros(2*T,1) ;
h=[A ; B] \ [c ; d] ;
dz(1:2*n)=h ; % da dy(1:n)
a=h(1:n) ; y=h(n+1:2*n) ;

% backsubstitute to get the rest of the variables
dz(n+1+n:T+n)=P(n+1:T,:)*y+M(n+1:T,:)*a+g(n+1:T) ;       % dy(n+1:T)
dz(T+n+1:2*T)=Q*y+R*a+r ; % lambda(n+1:T)

% ------------------------------------------------------------------
function dz=slow_newton(z,f,n,T)
% Performs slow Newton step computation, for test purposes
% It should give exactly the same result as newton_step

J=zeros(2*T,2*T) ;
y=z(n+1:n+T) ;
a=z(1:n) ;
lambda=[zeros(n,1) ; z(T+n+1:2*T) ; zeros(n,1)] ;

for i=1:T,
  J(i,1:n)=-lambda(i+1:i+n)' ;
  J(i,i+n)=2 ;
end ;  
for i=n+1:T,
  J(i-n:i,i+T)=[-a(n:-1:1) ; 1] ;
end ;  
for i=1:n,
  J(i+T,2*n+1-i:n+T-i)=-lambda(n+1:T)' ;
  J(i+T,n+T+1:2*T)=-y(n-i+1:T-i)' ;
end ;  
for i=n+1:T,
  J(i+T,i:i+n)=[-a(n:-1:1)' 1] ;
  J(i+T,1:n)=-y(i-1:-1:i-n)' ;
end ;  

dz=J\f ;
% ------------------------------------------------------------------
function y=estimate_y(a,x,n,T) 
% generate P such that 
% P*[y(1)..y(n)]'=[y(1)..y(T)]'
P=[eye(n) ; zeros(T-n,n) ] ;
for t=n+1:T,
  P(t,:)=a(n:-1:1)'*P(t-n:t-1,:) ;
end ;  
% solve the LS problem for y(1)..y(n)
y=P\x ;
y=P*y ; % initial guess of y

% ------------------------------------------------------------------
function lambda=estimate_lambda(a,x,y,n,T)
% initial guess of lambda - not very good
% generate Q,ee,c from Ib such that 
% Q*[lambda(T-n+1)..lambda(T)]'+c=[lambda(n+1)..lambda(T)]'
Q=zeros(T-n,n) ;
c=zeros(T-n,1) ;
Q(T-n+1-n:T-n,1:n)=eye(n) ;

for i=T-n:-1:n+1,
  c(i-n)=2*(x(i)-y(i)) ;
  Q(i-n,:)=a'*Q(i-n+1:i,:) ;
  c(i-n,:)=c(i-n)+a'*c(i-n+1:i) ;
end ; 
  
% transform the equations I and II to 
% R*[lambda(T-n+1)..lambda(T)]'=d
R=zeros(T+n,n) ; d=zeros(T+n,1) ;
for i=1:n,
  R(i,:)=-a(n-i+1:n)'*Q(1:i,:) ;
  d(i,:)=a(n-i+1:n)'*c(1:i,:)+2*(x(i)-y(i)) ;
end ;  
ae=[1 -a'] ;
for i=n+1:T,
  k=min(i+n,T)-i+1 ;
  R(i,:)=ae(1:k)*Q(i-n:i-n+k-1,:) ;
  d(i,:)=-ae(1:k)*c(i-n:i-n+k-1)+2*(x(i)-y(i)) ;
end ;  
for i=T+1:T+n,
  s=T+n-i+1 ;
  R(i,:)=-y(s:s+T-n-1)'*Q ;
  d(i,:)=y(s:s+T-n-1)'*c ;
end ;  

le=R\d ;
lambda=Q*le+c ;
% ------------------------------------------------------------------
function lgrad=lambda_grad(z,n,T,x,f)
% gradient of the Q=sum(f.^2) criterium with respect to lambda(n+1:T)
y=z(n+1:n+T) ;
a=z(1:n) ;

lgrad=zeros(T-n,1) ;
ae=[1 -a'] ; ae=ae(n+1:-1:1) ;
for i=1:T-n ;
  t=i+n ;
  lgrad(i)=2*(ae*f(i:t)-f(T+1:T+n)'*y(t-1:-1:i)) ;
end ;
% ------------------------------------------------------------------
function alpha=get_alpha(z,n,T,x,f,lgrad)
y=z(n+1:n+T) ;
a=z(1:n) ;
elgrad=[zeros(n,1) ; lgrad ; zeros(n,1) ] ;
r=zeros(T,1) ;
left=0 ;
for t=1:T,
  r(t)=[1 -a']*elgrad(t:t+n) ;
  left=left-f(t)*r(t) ;
end ;  
for i=1:n,
  s(i)=lgrad'*y(n+1-i:T-i) ;
  left=left+f(T+i)*s(i) ;
end ;  
right=sum(r.^2)+sum(s.^2) ;
alpha=left/right ;

% ------------------------------------------------------------------
function newz=estimate_lambda2(z,n,T,x) 
y=z(n+1:n+T) ;
a=z(1:n) ;

f=eval_right_side(z,n,T,x) ;
sum(f.*f) 
for i=1:10,
lgrad=lambda_grad(z,n,T,x,f) ;
alpha=get_alpha(z,n,T,x,f,lgrad)
z(T+n+1:2*T)=z(T+n+1:2*T)+0.5*lgrad ;
end ;

newz=z ;
