function [a,y]=tstls4(x,n) 
% Toeplitz Structured Total Least Squares Algorithm
% 'x' containts measurements corrupted by white noise
% on output 'a' contains the coefficients of an AR system of the order 'n'
% generating output 'y', while (x-y) is minimised in Euclidean sense
% y(i)=a(1)*y(i-1)+a(2)*y(i-2)+...+a(n)*y(i-n)
% a=arg min sum((y(i)-x(i))^2)
%
% Use quasi Newton iterative method to solve Lagrange equations
% It should have quadratic convergence.
% The implementation tries to be quick but ignores some numerical issues on
% the premises of a restricted magnitude range of the quantities involved
%
% This version tries to find minimum on backtracking
% It also tries to find a good initial estimate
% Some additional heuristic is applied to improve the convergence
%
% Jan Kybic, 1997
%
% Usage: [a y]=tstls4(x,n)

x=x(:) ;      % make x a column vector
T=length(x) ; % number of available samples
m=T-n ;       % number of equations

prec=1e-6 ;   % desired precision
maxiter=100 ;  % maximum number of iterations
maxiniter=10 ;% maximum inner number of iterations

% initial guess of a
 a=burg1(x,n)' ;

% scale the problem
scale=sum(abs(x))/T ;
% scale=1 ;
x=x/scale ;

y=estimate_y(a,x,n,T) ;
% y=x ;
% [a y]=tstlsw(x,n,y,a) ;
% [a y]=tstlsw(x,n) ;

z=[a ; y ; zeros(T-n,1)];

% z=estimate_lambda2(z,n,T,x) ;

% initial value of z= [ a ; y ; lambda(n+1:T)]
% z=[a ; y ; lambda];
% z=[a ; y ; 0*ones(T-n,1)];

fval=0 ;
% constants for the backtracking
alphamin=0.01 ;
alphamax=0.5 ;
alphafirst=0.38197 ; % golden ratio
% alphafirst=0.01 ;
sml=1e-4 ;
tol=1e-4 ;
abstol=1e-20 ;
maxiniter=20 ;
gld=0.38197 ; % 1-golden ratio
i=0 ; % number of evaluations
dz=zeros(2*T,1) ;
while i<maxiter,
  
  % make full step first
  f=eval_right_side(z+dz,n,T,x) ;
  i=i+1 ;
  oldfval=fval ; fval=sum(f.*f) ;
  
  if i>1,
    J0=oldfval ; J1=fval ;
    alpha0=0 ; alpha1=1 ;
    stps=1 ;
    if J1<J0,
      J=J0 ;
      % the function seems to decrease, proceed
      while J1<J, 
	J=J1 ; alpha=alpha1 ;
	alpha1=10*alpha1 ;
	f=eval_right_side(z+dz*alpha1,n,T,x) ; 
	J1=sum(f.^2) ;
	stps=stps+1 ;
      end ;
    else
      % the function increases, go bacward until it decreases
      alpha=0.5 ;
      f=eval_right_side(z+dz*alpha,n,T,x) ; 
      J=sum(f.^2) ;
      while J>J0,
	J1=J ; alpha1=alpha ;
	alpha=0.1*alpha1 ;
	f=eval_right_side(z+dz*alpha,n,T,x) ; 
	J=sum(f.^2) ;
	stps=stps+1 ;
      end ;  
    end ;
    % now the minimum is bracketed between alpha0,alpha, alpha1
    for j=1:maxiniter,
      if j>1 & abs(alpha1-alpha0)<max([tol*alpha abstol]), % success
	z=z+dz*alpha ;
	f=newf ; fval=newJ ;
	disp(['Convergence in ' num2str(j) ' + ' num2str(stps) ' steps.']) ;
	break ;
      end ;
      % find new alpha by parabolic approximation
      dif1=alpha-alpha1 ; dif2=J-J0 ; dif3=alpha-alpha0 ; dif4=J-J1 ;
      den=dif1*dif2-dif3*dif4 ;
      if abs(den)<eps, % colinear points
	% use golden search
	if J1>J0,
	  newalpha=alpha*(1-gld)+gld*alpha0 ;
	else 
	  newalpha=alpha1*gld+(1-gld)*alpha ;
	end ;
      else
	% use parabolic approximation
	newalpha=alpha-0.5*(dif1^2*dif2-dif3^2*dif4)/den ;
	% force newalpha into bounds
	newalpha=max([min([newalpha alpha1-sml*(alpha1-alpha0)]) ...
	      sml*(alpha1-alpha0)+alpha0]) ;
      end ;  
      % we have the new alpha now
      newf=eval_right_side(z+dz*newalpha,n,T,x) ; 
      newJ=sum(f.^2) ;
      if newJ>J, % something's gone wrong - unsuccessful estimate
	if newalpha<alpha,
	  alpha0=newalpha ; J0=newJ ;
	else
	  alpha1=newalpha ; J1=newJ ;
	end ;
      else
	if newalpha<alpha ;
	  alpha1=alpha ; J1=J ; alpha=newalpha ; J=newJ ; f=newf ;
	else
	  alpha0=alpha ; J0=J ; alpha=newalpha ; J=newJ ; f=newf ;
	end ;
      end ;
    end ; % of for j
  end ;  % of if i>1  
  
  disp(['Evaluation(main) ' num2str(i) ' fval=' num2str(fval)]) ;
  if fval<prec, % desired precision reached ? 
    break ;
  end ;
  
  % dz=-slow_newton(z,f,n,T) ; % this should give the same results
  dz=-newton_step(z,f,n,T) ; 
end ;

% all the maxiter iterations taken or convergence attained
a=z(1:n) ;
y=z(n+1:T+n)*scale ;

% --------------------------------------------------------------------
function f=eval_right_side(z,n,T,x)
% Evaluates the Lagrangian f=F(z) at point z
% z=[y(1)..y(T) a(1)..a(n) lambda(n+1)..lamda(T)] is a column vector
% f has the dimension 2*T x 1
f=zeros(2*T,1) ;
y=z(n+1:n+T) ;
a=z(1:n) ;
lambda=[zeros(n,1) ; z(T+n+1:2*T) ; zeros(n,1)] ;
% equations I
for t=1:T,
  f(t)=2*(y(t)-x(t)) + [1 -a(1:n)']*lambda(t:t+n) ;
end ;
% equations II
for k=1:n,
  f(T+k)= - lambda(n+1:T)' * y(n+1-k:T-k) ;
end ;  
% equations III
for t=n+1:T,
  f(T+t)=y(t)-a(1:n)'*y(t-1:-1:t-n) ;
end ;  

% --------------------------------------------------------------------
function dz=newton_step(z,f,n,T) 
% given the point z and the corresponding Lagrangian f=F(z),
% find the Newton step dz=inv(dF/dz)*f taking advantage of the 
% Jacobian (dF/dz) matrix sparsity and structure
y=z(n+1:n+T) ;
a=z(1:n) ;
lambda=[zeros(n,1) ; z(T+n+1:2*T)] ;

% generate P,M,g from III such that 
% P*[dy(1)..dy(n)]'+M*[da(1)..da(n)]'+g=[dy(1)..dy(T)]'
P=[eye(n) ; zeros(T-n,n) ] ;
M=zeros(T,n) ;
g=zeros(T,1) ;
for t=n+1:T,
  P(t,:)=a(n:-1:1)'*P(t-n:t-1,:) ;
  g(t)=f(t+T)+a(n:-1:1)'*g(t-n:t-1) ;
  M(t,:)=y(t-1:-1:t-n)'+a(n:-1:1)'*M(t-n:t-1,:) ;
end ;  

% generate (from Ib) Q,R,r such that 
% Q*[dy(1)..dy(n)]'+R*[da(1)..da(n)]'+r=[dlambda(n+1)..dlambda(T)]'
Q=zeros(T-n,n) ;
r=zeros(T-n,1) ;
Q(T-n,:)=-2*P(T,:) ;
r(T-n)=f(T)-2*g(T) ;
R=zeros(T-n,n) ;
R(T-n,:)=-2*M(T,:) ;
for t=T-1:-1:n+1 ;
  k=min(T-t,n) ;
  Q(t-n,:)=-2*P(t,:)+a(1:k)'*Q(t-n+1:t-n+k,:) ;
  r(t-n)=f(t)-2*g(t)+a(1:k)'*r(t-n+1:t-n+k) ;
  mx=min(t+n,T) ;
  R(t-n,1:mx-t)=lambda(t+1:mx)' ;
  R(t-n,:)=R(t-n,:)-2*M(t,:)+a(1:k)'*R(t-n+1:t-n+k,:) ;
end ;  

% convert equations II to a form A*[a(1)..a(n) y(1)..y(n)]'=c
A=zeros(n,2*n) ;
c=zeros(n,1) ;
for i=1:n,
  c(i)=f(i+T)+lambda(n+1:T)'*g(n-i+1:T-i) ;
  c(i)=c(i)+y(n+1-i:T-i)'*r(1:T-n) ;
  A(i,1:n)=-y(n+1-i:T-i)'*R-lambda(n+1:T)'*M(n-i+1:T-i,:) ;
  A(i,n+1:2*n)=-y(n+1-i:T-i)'*Q-lambda(n+1:T)'*P(n-i+1:T-i,:) ;
end ;  

% convert equations Ia to B*[a(1)..a(n) y(1)..y(n)]'=d
B=zeros(n,2*n) ;
d=zeros(n,1) ;
for i=1:n,
  B(i,1:n)=-lambda(i+1:i+n)'-a(n-i+1:n)'*R(1:i,:) ;
  B(i,n+1:2*n)=2*P(i,:)-a(n-i+1:n)'*Q(1:i,:) ;
  d(i)=f(i)-2*g(i)+a(n-i+1:n)'*r(1:i) ;
end ;

% solve
dz=zeros(2*T,1) ;
h=[A ; B] \ [c ; d] ;
dz(1:2*n)=h ; % da dy(1:n)
a=h(1:n) ; y=h(n+1:2*n) ;

% backsubstitute to get the rest of the variables
dz(n+1+n:T+n)=P(n+1:T,:)*y+M(n+1:T,:)*a+g(n+1:T) ;       % dy(n+1:T)
dz(T+n+1:2*T)=Q*y+R*a+r ; % lambda(n+1:T)

% ------------------------------------------------------------------
function dz=slow_newton(z,f,n,T)
% Performs slow Newton step computation, for test purposes
% It should give exactly the same result as newton_step

J=zeros(2*T,2*T) ;
y=z(n+1:n+T) ;
a=z(1:n) ;
lambda=[zeros(n,1) ; z(T+n+1:2*T) ; zeros(n,1)] ;

for i=1:T,
  J(i,1:n)=-lambda(i+1:i+n)' ;
  J(i,i+n)=2 ;
end ;  
for i=n+1:T,
  J(i-n:i,i+T)=[-a(n:-1:1) ; 1] ;
end ;  
for i=1:n,
  J(i+T,2*n+1-i:n+T-i)=-lambda(n+1:T)' ;
  J(i+T,n+T+1:2*T)=-y(n-i+1:T-i)' ;
end ;  
for i=n+1:T,
  J(i+T,i:i+n)=[-a(n:-1:1)' 1] ;
  J(i+T,1:n)=-y(i-1:-1:i-n)' ;
end ;  

dz=J\f ;
% ------------------------------------------------------------------
function y=estimate_y(a,x,n,T) 
% generate P such that 
% P*[y(1)..y(n)]'=[y(1)..y(T)]'
P=[eye(n) ; zeros(T-n,n) ] ;
for t=n+1:T,
  P(t,:)=a(n:-1:1)'*P(t-n:t-1,:) ;
end ;  
% solve the LS problem for y(1)..y(n)
y=P\x ;
y=P*y ; % initial guess of y

% ------------------------------------------------------------------
function lambda=estimate_lambda(a,x,y,n,T)
% initial guess of lambda - not very good
% generate Q,ee,c from Ib such that 
% Q*[lambda(T-n+1)..lambda(T)]'+c=[lambda(n+1)..lambda(T)]'
Q=zeros(T-n,n) ;
c=zeros(T-n,1) ;
Q(T-n+1-n:T-n,1:n)=eye(n) ;

for i=T-n:-1:n+1,
  c(i-n)=2*(x(i)-y(i)) ;
  Q(i-n,:)=a'*Q(i-n+1:i,:) ;
  c(i-n,:)=c(i-n)+a'*c(i-n+1:i) ;
end ; 

% transform the equations I and II to 
% R*[lambda(T-n+1)..lambda(T)]'=d
R=zeros(T+n,n) ; d=zeros(T+n,1) ;
for i=1:n,
  R(i,:)=-a(n-i+1:n)'*Q(1:i,:) ;
  d(i,:)=a(n-i+1:n)'*c(1:i,:)+2*(x(i)-y(i)) ;
end ;  
ae=[1 -a'] ;
for i=n+1:T,
  k=min(i+n,T)-i+1 ;
  R(i,:)=ae(1:k)*Q(i-n:i-n+k-1,:) ;
  d(i,:)=-ae(1:k)*c(i-n:i-n+k-1)+2*(x(i)-y(i)) ;
end ;  
for i=T+1:T+n,
  s=T+n-i+1 ;
  R(i,:)=-y(s:s+T-n-1)'*Q ;
  d(i,:)=y(s:s+T-n-1)'*c ;
end ;  

le=R\d ;
lambda=Q*le+c ;
% ------------------------------------------------------------------
function lgrad=lambda_grad(z,n,T,x,f)
% gradient of the Q=sum(f.^2) criterium with respect to lambda(n+1:T)
y=z(n+1:n+T) ;
a=z(1:n) ;

lgrad=zeros(T-n,1) ;
ae=[1 -a'] ; ae=ae(n+1:-1:1) ;
for i=1:T-n ;
  t=i+n ;
  lgrad(i)=2*(ae*f(i:t)-f(T+1:T+n)'*y(t-1:-1:i)) ;
end ;
% ------------------------------------------------------------------
function alpha=get_alpha(z,n,T,x,f,lgrad)
y=z(n+1:n+T) ;
a=z(1:n) ;
elgrad=[zeros(n,1) ; lgrad ; zeros(n,1) ] ;
r=zeros(T,1) ;
left=0 ;
for t=1:T,
  r(t)=[1 -a']*elgrad(t:t+n) ;
  left=left-f(t)*r(t) ;
end ;  
for i=1:n,
  s(i)=lgrad'*y(n+1-i:T-i) ;
  left=left+f(T+i)*s(i) ;
end ;  
right=sum(r.^2)+sum(s.^2) ;
alpha=left/right ;

% ------------------------------------------------------------------
function newz=estimate_lambda2(z,n,T,x) 
y=z(n+1:n+T) ;
a=z(1:n) ;

f=eval_right_side(z,n,T,x) ;
sum(f.*f) 
for i=1:10,
  lgrad=lambda_grad(z,n,T,x,f) ;
  alpha=get_alpha(z,n,T,x,f,lgrad)
  z(T+n+1:2*T)=z(T+n+1:2*T)+0.5*lgrad ;

  for j=1:100,
    z(T+n+1:2*T)=z(T+n+1:2*T)-0.01*lgrad ;
    f=eval_right_side(z,n,T,x) ;
    q(j)=sum(f.*f) ;
  end ;
  plot(q)
  return
end ;

newz=z ;
