function [a,y]=tstls5(x,n) 
% Toeplitz Structured Total Least Squares Algorithm
% 'x' containts measurements corrupted by white noise
% on output 'a' contains the coefficients of an AR system of the order 'n'
% generating output 'y', while (x-y) is minimised in Euclidean sense
% y(i)=a(1)*y(i-1)+a(2)*y(i-2)+...+a(n)*y(i-n)
% a=arg min sum((y(i)-x(i))^2)
%
% Based on STLNB algorithm from S. Van Huffel, Haesun Park, J. Ben Rosen: 
% 'Formulation and Solution of Structured Total Least Norm Problems for
% Parameter Estimation', IEEE Transactions on Signal Processing, 1996
%
% Fast method, diverges for wrong initial estimates
%
% Jan Kybic, 1997
%
% Usage: [a y]=tstls5(x,n)


x=x(:) ;      % make x a column vector
T=length(x) ; % number of available samples
m=T-n ;       % number of equations

prec=1e-6 ;   % desired precision
maxiter=50 ; % maximum number of iterations

% construct the left hand side matrix
A=zeros(m,n) ;
for i=1:n, A(:,i)=x(i:i+m-1) ; end ;

b=x(n+1:T) ; % right hand side

% initialize the iterative procedure
a=A\b ;
a=a(n:-1:1) ;
z=[x ; a] ;

for i=1:maxiter,
  % z is the current point we want to improve
  J=crit(z,T,x) 
  [C o]=calc_base(z,T,n) ;
  s=[x-z(1:T) ; zeros(n,1)] ; % idealni smer
  dz=C'*C*(s-o)+o ;
  z=z+dz ;
end ;

y=z(1:T) ;
a=z(T+1:T+n)' ;


% -------------------------------------------------
function J=crit(z,T,x)
% calculate the criteria
J=sum((z(1:T)-x).^2) ;

% -------------------------------------------------
function dz=complete(dz,z,T,n)
% given first n and last n values of dz, complete the rest
y=z(1:T) ;
a=z(T+1:T+n) ;
dy=[ dz(1:n) ; zeros(T-n,1)] ;
da=dz(T+1:T+n) ;

% calculate the right side
r=zeros(T,1) ;
for i=n+1:T,
  r(i)=[-a(n:-1:1)' 1]*y(i-n:i)-da(n:-1:1)'*y(i-n:i-1) ;
end ;
% calculate the remaining dy
for i=n+1:T,
  dy(i)=-r(i)+a(n:-1:1)'*dy(i-n:i-1) ;
end ;  
dz=[dy ; da] ;

% -------------------------------------------------
function [C,o]=calc_base(z,T,n)
% return the orthonormal base C of the linearized constraint subspace
% and vector o that lies in that subspace 
E=eye(2*n) ;
C=[ E(:,1:n) zeros(2*n,T-n) E(:,n+1:2*n)];

o=complete(zeros(T+n,1),z,T,n) ;
for i=1:2*n,
  C(i,:)=complete(C(i,:)',z,T,n)' ;  
  C(i,:)=C(i,:)-o' ;
end ;  

for i=1:2*n,
  for j=1:i-1,
    g=C(i,:)*C(j,:)' ;
    C(i,:)=C(i,:)-g*C(j,:) ;
  end ;  
  g=C(i,:)*C(i,:)' ;
  C(i,:)=C(i,:)/sqrt(g) ;
end ;  

C=C(1:2*n,:) ;


