function [a,y]=tstlsw(x,n,y,a) 
% Toeplitz Structured Total Least Squares Algorithm
% 'x' containts measurements corrupted by white noise
% on output 'a' contains the coefficients of an AR system of the order 'n'
% generating output 'y', while (x-y) is minimised in Euclidean sense
% y(i)=a(1)*y(i-1)+a(2)*y(i-2)+...+a(n)*y(i-n)
% a=arg min sum((y(i)-x(i))^2)
%
% Uses weighting to get unconstrained problem that solves using (conjugate) 
% gradient method
%
% The last two arguments are optional starting values
%
% Jan Kybic, 1997
%
% Usage: [a y]=tstlsw(x,n,[y,a])

x=x(:) ;      % make x a column vector
T=length(x) ; % number of available samples
m=T-n ;       % number of equations

prec=1e-6 ;   % desired overall relative precision
maxiter=100 ;  % maximum number of iterations
miniter=5 ;    % minimum number of iterations
omega=1e2 ;   % weight

if nargin<3,  % estimate
  a=burg1(x,n)' ;
  y=x ;
  % y=estimate_y(a,x,n,T) ;
end ;  

% column state vector z=[y n]
z=[y ; a] ;

for i=1:maxiter,
[J0,gr]=crit_and_grad(z,n,T,x,omega) ; % J0 

if i==1,
  h=gr ; g=gr ;
else
  gamma=(gr-g)'*gr/(g'*g) ;
  g=gr ; h=gamma*h+g ;
end ;

[z J]=make_step(J0,-h,z,n,T,x,omega) ; 
disp(['Iteration ' num2str(i) ' J=' num2str(J)]) ;
% a=z(T+1:T+n)' 
% sqrt(sum((a-a1).^2)) 
if i>miniter & (J0-J)/J<prec,  break ; end ;
end ;

y=z(1:T)' ;
a=z(T+1:T+n)' ;

% ---------------------------------------------------------------
function J=crit(z,n,T,x,omega) 
% evaluates the criterium J
J=sum((z(1:T)-x).^2) ;
J1=0 ;
gr=zeros(T+n,1) ;
aer=[-z(T+n:-1:T+1)' 1] ;
for i=n+1:T,
  ee=aer*z(i-n:i) ;
  J1=J1+ee^2 ;
end ;  
J=J1*omega+J ;
% ---------------------------------------------------------------
function [J,gr]=crit_and_grad(z,n,T,x,omega) 
% evaluates the criterium J and gradient gr
J=sum((z(1:T)-x).^2) ;
J1=0 ;
ee=zeros(T,1) ;
gr=zeros(T+n,1) ;
aer=[-z(T+n:-1:T+1)' 1] ;
for i=n+1:T,
  ee(i)=aer*z(i-n:i) ;
  J1=J1+sum(ee(i).^2) ;
end ;  
J=J1*omega+J ;

ae=[1 -z(T+1:T+n)'] ;
eee=[ee ; zeros(n,1) ] ;
for i=1:T,
  gr(i)=2*(z(i)-x(i))+2*omega*ae*eee(i:i+n) ;
end ;  
for i=1:n,
  gr(i+T)=-2*omega*ee(n+1:T)'*z(n+1-i:T-i) ;
end ;  

% ---------------------------------------------------------------
function [newz,newJ]=make_step(J0,h,z,n,T,x,omega,maxiniter)
% This function find minimum of J along the direction h
% using simplified Brent's method

maxiniter=50 ;% maximum number of inner loop iterations
sml=1e-4 ;    % small number :-)
gld=0.38197 ; % 1-golden ratio
tol=1e-4 ;    % relative error
abstol=1e-20 ; % absolute error

alpha0=0 ;

magh=sqrt(sum(h.^2))/10 ; % expect approximately unitary steps
if magh>1, h=h/magh ; end ;

% make full step first
J1=crit(z+h,n,T,x,omega) ;
stps=2 ;
alpha1=1 ;     % save the full step result
if J1<J0,
  J=J0 ;
  % the function seems to decrease, proceed
  while J1<J, 
    J=J1 ; alpha=alpha1 ;
    alpha1=10*alpha1 ;
    J1=crit(z+h*alpha1,n,T,x,omega) ; 
    stps=stps+1 ;
  end ;
else
  % the function increases, go bacward until it decreases
  alpha=0.5 ;
  J=crit(z+h*alpha,n,T,x,omega) ; 
  while J>J0,
    J1=J ; alpha1=alpha ;
    alpha=0.1*alpha1 ;
    J=crit(z+h*alpha,n,T,x,omega) ; 
    stps=stps+1 ;
  end ;  
end ;

% now the minimum is bracketed between alpha0,alpha, alpha1
for j=1:maxiniter,
  if j>1 & abs(alpha1-alpha0)<max([tol*alpha abstol]), % success
    newz=z+h*alpha ;
%    disp(['Convergence in ' num2str(j) ' + ' num2str(stps) ' steps. ' ...
%	  'alpha=' num2str(alpha)]) ;
    newJ=J ;
    return ;
  end ;
  % find new alpha by parabolic approximation
  dif1=alpha-alpha1 ; dif2=J-J0 ; dif3=alpha-alpha0 ; dif4=J-J1 ;
  den=dif1*dif2-dif3*dif4 ;
  if abs(den)<eps, % colinear points
    if J1>J0,
      newalpha=alpha*(1-gld)+gld*alpha0 ;
    else 
      newalpha=alpha1*gld+(1-gld)*alpha ;
    end ;
  else
    newalpha=alpha-0.5*(dif1^2*dif2-dif3^2*dif4)/den ;
    % force into bounds
    newalpha=max([min([newalpha alpha1-sml*(alpha1-alpha0)]) ...
	  sml*(alpha1-alpha0)+alpha0]) ;
  end ;  
  % we have the new alpha now
  newJ=crit(z+h*newalpha,n,T,x,omega) ; 

  if newJ>J, % something's gone wrong - unsuccessful estimate
    if newalpha<alpha,
      alpha0=newalpha ; J0=newJ ;
    else
      alpha1=newalpha ; J1=newJ ;
    end ;
  else
    if newalpha<alpha ;
      alpha1=alpha ; J1=J ; alpha=newalpha ; J=newJ ;
    else
      alpha0=alpha ; J0=J ; alpha=newalpha ; J=newJ ;
    end ;
  end ;
  
end ; % of for j

disp('Divergence in make_step') ;  
newz=z+h*alpha ;
newJ=J ;

% ------------------------------------------------------------------
function y=estimate_y(a,x,n,T) 
% generate P such that 
% P*[y(1)..y(n)]'=[y(1)..y(T)]'
P=[eye(n) ; zeros(T-n,n) ] ;
for t=n+1:T,
  P(t,:)=a(n:-1:1)'*P(t-n:t-1,:) ;
end ;  
% solve the LS problem for y(1)..y(n)
y=P\x ;
y=P*y ; % initial guess of y
